--  Clocking
--  This is very simple: the fpga has one main clock domain (100Mhz),
--  but needs to generate:
--  * 400Mhz for the oserdes to the system board
--  * 200Mhz for delay (?)
--
--  TBD:
--  The iserdes uses their own input clocks, which is syntonized on the main
--  clock (but the phase is not known).

library ieee;
use ieee.std_logic_1164.all;

library unisim;
use unisim.vcomponents.all;

entity fsi_adc_clk is
  port (
    --  Reset from backplane and powergood.
    fpga_reset_n_i : std_logic;

    --  Clock (100Mhz)
    fpga_clk_p_i : std_logic;
    fpga_clk_n_i : std_logic;

    clk_100_o : out std_logic;
    clk_400_o : out std_logic;
    rst_n_o : out std_logic
  );
end fsi_adc_clk;

architecture arch of fsi_adc_clk is
  signal clk_100_in, clk_100_out, clk_100_buf : std_logic;
  signal clk_400_out, clk_400_buf : std_logic;
  signal clk_200_out, clk_200_buf : std_logic;
  signal clk_fb_out, clk_fb_buf : std_logic;
  signal rst : std_logic;
  signal rst_out : std_logic;
  signal locked : std_logic;
  signal rdy : std_logic;
  signal idly_rst : std_logic;
  signal idly_rst_cnt : natural range 0 to 7;
begin
  inst_IBUFDS_clk : IBUFGDS
    port map (
      O  => clk_100_in,    -- Buffer output
      I  => fpga_clk_p_i,  -- Diff_p buffer input
      IB => fpga_clk_n_i); -- Diff_n buffer input

  rst <= not fpga_reset_n_i;

  --  VCO=1200Mhz
  inst_PLLE2 : PLLE2_BASE
    generic map (
      BANDWIDTH => "OPTIMIZED", -- OPTIMIZED, HIGH, LOW
      CLKFBOUT_MULT => 12,       -- Multiply value for all CLKOUT, (2-64)
      CLKFBOUT_PHASE => 0.0,   -- Phase offset in degrees of CLKFB, (-360.000-360.000).
      CLKIN1_PERIOD => 10.0,   -- Input clock period in ns to ps resolution
      CLKOUT0_DIVIDE => 12, -- Divide amount for each CLKOUT (1-128)
      CLKOUT1_DIVIDE => 3,
      CLKOUT2_DIVIDE => 6,
      CLKOUT3_DIVIDE => 1,
      CLKOUT4_DIVIDE => 1,
      CLKOUT5_DIVIDE => 1,
      CLKOUT0_DUTY_CYCLE => 0.5, -- Duty cycle for each CLKOUT (0.001-0.999).
      CLKOUT1_DUTY_CYCLE => 0.5,
      CLKOUT2_DUTY_CYCLE => 0.5,
      CLKOUT3_DUTY_CYCLE => 0.5,
      CLKOUT4_DUTY_CYCLE => 0.5,
      CLKOUT5_DUTY_CYCLE => 0.5,
      CLKOUT0_PHASE => 0.0, -- Phase offset for each CLKOUT (-360.000-360.000).
      CLKOUT1_PHASE => 0.0,
      CLKOUT2_PHASE => 0.0,
      CLKOUT3_PHASE => 0.0,
      CLKOUT4_PHASE => 0.0,
      CLKOUT5_PHASE => 0.0,
      DIVCLK_DIVIDE => 1,
      REF_JITTER1 => 0.0,
      STARTUP_WAIT => "FALSE"  -- Delay DONE until PLL Locks, ("TRUE"/"FALSE")
      )
    port map (
      -- Clock Outputs: 1-bit (each) output: User configurable clock outputs
      CLKOUT0 => clk_100_out,
      CLKOUT1 => clk_400_out,
      CLKOUT2 => clk_200_out,
      CLKOUT3 => open,
      CLKOUT4 => open,
      CLKOUT5 => open,

      CLKFBOUT => clk_fb_out, -- Feedback Clock
      LOCKED => locked,
      CLKIN1 => clk_100_in,
      PWRDWN => '0',
      RST => rst, -- 1-bit input: Reset
      CLKFBIN => clk_fb_buf
      );

  int_BUFG_100 : BUFG
    port map (
      O => clk_100_buf,
      I => clk_100_out
    );

  int_BUFG_400 : BUFG
    port map (
      O => clk_400_buf,
      I => clk_400_out
    );

  int_BUFG_200 : BUFG
    port map (
      O => clk_200_buf,
      I => clk_200_out
    );

  int_BUFG_fb : BUFG
    port map (
      O => clk_fb_buf,
      I => clk_fb_out
    );

  inst_IDELAYCTRL : IDELAYCTRL
    port map (
      RDY => rdy,       -- 1-bit output: Ready output
      REFCLK => clk_200_buf, -- 1-bit input: Reference clock input
      RST => idly_rst        -- 1-bit input: Active high reset input
    );

  --  Reset pulse for the idelayctrl.  At least 50ns
  process (clk_100_buf, rst, locked) is
  begin
    if rst = '1' or locked = '0' then
      idly_rst <= '1';
      idly_rst_cnt <= 7;
    elsif rising_edge (clk_100_buf) then
      if idly_rst_cnt = 0 then
        idly_rst <= '0';
      else
        idly_rst_cnt <= idly_rst_cnt - 1;
      end if;
    end if;
  end process;

  process (clk_100_buf, rst, locked, rdy) is
  begin
    if rst = '1' or locked = '0' or rdy = '0' then
      rst_out <= '1';
    elsif rising_edge (clk_100_buf) then
      rst_out <= '0';
    end if;
  end process;

  rst_n_o <= not rst_out;
  clk_100_o <= clk_100_buf;
  clk_400_o <= clk_400_buf;
end arch;
