library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library unisim;
use unisim.vcomponents.all;

use work.fsi_adc_pkg.all;

entity fsi_adc_rx is
  port (
    clk_100_i : std_logic;
    rst_n_i : std_logic;

    --  ADC data
    adc_outa_p_i : std_logic_vector(7 downto 0);
    adc_outa_n_i : std_logic_vector(7 downto 0);
    adc_outb_p_i : std_logic_vector(7 downto 0);
    adc_outb_n_i : std_logic_vector(7 downto 0);

    --  ADC bit clock
    adc_lclk_p_i : std_logic;
    adc_lclk_n_i : std_logic;

    --  ADC frame
    adc_adclk_p_i : std_logic;
    adc_adclk_n_i : std_logic;

    locked_o : out std_logic;
    data_o : out t_slv_8x12;
    dbg_slip : out std_logic;
    dbg_f : out std_logic_vector(5 downto 0);
    dbg_0a : out std_logic_vector(5 downto 0);
    dbg_0b : out std_logic_vector(5 downto 0);
    dbg_clk : out std_logic
  );
end fsi_adc_rx;

architecture arch of fsi_adc_rx is
    --  Inputs.  0:outa(0), 1:outb(0), 2:outa(1), ...
    signal din : std_logic_vector(8*2 + 1 - 1 downto 0);
    type t_slv6 is array (natural range <>) of std_logic_vector(5 downto 0);
    signal dout : t_slv6(din'range);

    signal lclk_in, lclk_io, lclk_div : std_logic;
    signal lclk_io_n : std_logic;

    signal rst : std_logic;
    signal bitslip : std_logic;
    signal lock_cnt : unsigned(4 downto 0);
    signal unlock_cnt : unsigned(3 downto 0);
    signal locked : std_logic;

    signal fifo_din, fifo_dout : std_logic_vector (8*12 - 1 downto 0);
    signal fifo_rd, fifo_empty : std_logic;
begin
    --  Input pads
    --  Polarity is inverted on OUT lines
    gen_ibuf: for i in 7 downto 0 generate
        inst_ibuf_outa: IBUFDS_DIFF_OUT
            port map (
                OB => din (2*i + 0),
                I => adc_outa_p_i (i),
                IB => adc_outa_n_i (i)
            );
        inst_ibuf_outb: IBUFDS_DIFF_OUT
            port map (
                OB => din (2*i + 1),
                I => adc_outb_p_i (i),
                IB => adc_outb_n_i (i)
            );
    end generate;

    inst_ibuf_frame: IBUFDS
        port map (
            O => din (16),
            I => adc_adclk_p_i,
            IB => adc_adclk_n_i
        );

    inst_ibuf_lclk: IBUFGDS
        port map (
            O => lclk_in,
            I => adc_lclk_p_i,
            IB => adc_lclk_n_i
        );
    
    --  Clocking
    --  According to ug471 (p153), ISERDES2 works with BUFIO + BUFR (/3)
    inst_bufio_lclk: BUFIO
        port map (
            O => lclk_io,
            I => lclk_in
        );

    lclk_io_n <= not lclk_io;

    inst_bufr_lclk: BUFR
        generic map (
            BUFR_DIVIDE => "3", -- Values: "BYPASS, 1, 2, 3, 4, 5, 6, 7, 8"
            SIM_DEVICE => "7SERIES" -- Must be set to "7SERIES"
        )
        port map (
            O => lclk_div, -- 1-bit output: Clock output port
            CE => '1',     -- 1-bit input: Active high, clock enable (Divided modes only)
            CLR => '0',    -- 1-bit input: Active high, asynchronous clear (Divided modes only)
            I => lclk_in   -- 1-bit input: Clock buffer input driven by an IBUF, MMCM or local interconnect
        );

    --  ISERDES
    rst <= not rst_n_i;

    gen_iserdes: for i in din'range generate
        inst_ISERDESE2 : ISERDESE2
            generic map (
                DATA_RATE => "DDR", -- DDR, SDR
                DATA_WIDTH => 6,   -- Parallel data width (2-8,10,14)
                DYN_CLKDIV_INV_EN => "FALSE", -- Enable DYNCLKDIVINVSEL inversion (FALSE, TRUE)
                DYN_CLK_INV_EN => "FALSE", -- Enable DYNCLKINVSEL inversion (FALSE, TRUE)
                INIT_Q1 => '0',-- INIT_Q1 - INIT_Q4: Initial value on the Q outputs (0/1)
                INIT_Q2 => '0',
                INIT_Q3 => '0',
                INIT_Q4 => '0',
                INTERFACE_TYPE => "NETWORKING", -- MEMORY, MEMORY_DDR3, MEMORY_QDR, NETWORKING, OVERSAMPLE
                IOBDELAY => "NONE", -- NONE, BOTH, IBUF, IFD
                NUM_CE => 2,        -- Number of clock enables (1,2)
                OFB_USED => "FALSE",-- Select OFB path (FALSE, TRUE)
                SERDES_MODE => "MASTER",    -- MASTER, SLAVE

                SRVAL_Q1 => '0',    -- SRVAL_Q1 - SRVAL_Q4: Q output values when SR is used (0/1)
                SRVAL_Q2 => '0',
                SRVAL_Q3 => '0',
                SRVAL_Q4 => '0'
            )
            port map (
                O => open,  -- 1-bit output: Combinatorial output

                Q1 => dout(i)(5),   -- Q1 - Q8: 1-bit (each) output: Registered data outputs
                Q2 => dout(i)(4),
                Q3 => dout(i)(3),
                Q4 => dout(i)(2),
                Q5 => dout(i)(1),
                Q6 => dout(i)(0),
                Q7 => open,
                Q8 => open,

                SHIFTOUT1 => open,-- SHIFTOUT1, SHIFTOUT2: 1-bit (each) output: Data width expansion output ports
                SHIFTOUT2 => open,
                BITSLIP => bitslip,

                CE1 => '1',-- CE1, CE2: 1-bit (each) input: Data register clock enable inputs
                CE2 => '1',
                CLKDIVP => '0', -- 1-bit input: TBD

                CLK => lclk_io, -- 1-bit input: High-speed clock
                CLKB => lclk_io_n,   -- 1-bit input: High-speed secondary clock
                CLKDIV => lclk_div,  -- 1-bit input: Divided clock
                OCLK => '0',         -- 1-bit input: High speed output clock used when INTERFACE_TYPE="MEMORY"
                DYNCLKDIVSEL => '0', -- 1-bit input: Dynamic CLKDIV inversion
                DYNCLKSEL => '0',    -- 1-bit input: Dynamic CLK/CLKB inversion
                D => din(i),         -- 1-bit input: Data input
                DDLY => '0',         -- 1-bit input: Serial data from IDELAYE2
                OFB => '0',          -- 1-bit input: Data feedback from OSERDESE2
                OCLKB => '0',        -- 1-bit input: High speed negative edge output clock
                RST => rst,

                SHIFTIN1 => '0',-- SHIFTIN1, SHIFTIN2: 1-bit (each) input: Data width expansion input ports
                SHIFTIN2 => '0'
            );
    end generate;

    dbg_f <= dout(16);
    dbg_0a <= dout(0);
    dbg_0b <= dout(1);
    dbg_slip <= bitslip;
    dbg_clk <= lclk_div;

    --  Bitslip control.
    process (lclk_div) is
    begin
        if rising_edge(lclk_div) then
            if rst = '1' then
                --  Not locked
                lock_cnt  <= (others => '1');
                unlock_cnt  <= (others => '1');
                bitslip <= '0';
            else
                bitslip <= '0';
                -- Note: in the datasheet, the text is correct but diagrams p77-79 are not.
                -- TODO: check for alternate 1/0 ?
                if dout(16) = "111111" or dout(16) = "000000" then
                    --  Pattern is correct
                    --  If locked for enough cycles, set the 'locked' register.
                    if lock_cnt = 0 then
                        locked <= '1';
                    else
                        lock_cnt <= lock_cnt - 1;
                    end if;
                else
                    --  Wrong pattern, not locked
                    lock_cnt  <= (others => '1');
                    locked <= '0';

                    --  Slip bit once every N cycles.
                    if unlock_cnt = 0 then
                        bitslip <= '1';
                    end if;
                    unlock_cnt <= unlock_cnt - 1;
                end if;
            end if;
        end if;
    end process;

    --  Synchronizer for locked.
    inst_sync_locked: entity work.gc_sync
      generic map (
        g_sync_edge => "positive"
      )
      port map (
        clk_i => clk_100_i,
        rst_n_a_i => rst_n_i,
        d_i => locked,
        q_o => locked_o
      );
    
    --  Resync.
    --  All the serdes was done in the adc clock domain, but the
    --  fpga uses a different clock domain.
    --  The frequencies are the same (same source clock), but the
    --  phase is unknown.
    --  Use a simple FIFO to resync.
    --  TODO: that's simple but the latency is not well known.
    --  Possible strategy: register the output either on the
    --    rising edge or the falling edge.
    --    Choose the edge by oversampling with the 400Mhz clock.
    gen_fifo_fmt_in: for i in 7 downto 0 generate
        fifo_din(i * 12 + 11 downto i * 12 + 6) <= dout (2 * i + 0);
        fifo_din(i * 12 + 5  downto i * 12 + 0) <= dout (2 * i + 1);
    end generate;

    inst_fifo: entity work.generic_async_fifo
      generic map (
          g_data_width => fifo_din'length,
          g_size       => 16,
          g_with_rd_empty => true,
          g_with_wr_full => false,
          g_almost_full_threshold => 2,
          g_almost_empty_threshold => 2
      )
      port map (
          rst_n_i => rst_n_i,

          clk_wr_i => lclk_div,
          d_i => fifo_din,
          we_i => '1',

          clk_rd_i => clk_100_i,
          q_o => fifo_dout,
          rd_i => fifo_rd,
          rd_empty_o => fifo_empty
      );

    fifo_rd <= not fifo_empty;

    gen_fifo_fmt_out: for i in 7 downto 0 generate
        data_o(i) <= fifo_dout(i * 12 + 11 downto i * 12);
    end generate;
end arch;
