library ieee;
use ieee.std_logic_1164.all;

entity fsi_adc_spi is
  port (
    clk_100_i   :     std_logic;
    rst_n_i     :     std_logic;

    addr_i      :     std_logic_vector(7 downto 0);
    data_i      :     std_logic_vector(15 downto 0);
    data_o      : out std_logic_vector(15 downto 0);

    --  Pulse to start a transaction.
    start_i     :     std_logic;
    --  True while a transaction is in progress.
    busy_o      : out std_logic;

    --  Connections to the ADC
    adc_sclk_o  : out std_logic;
    adc_sdata_o : out std_logic;
    adc_sdout_i : in  std_logic;
    adc_cs_n_o  : out std_logic
  );
end fsi_adc_spi;

architecture arch of fsi_adc_spi is
  type t_state is (IDLE, ADDR, DATA, EXTRA);
  signal state     : t_state;

  signal din, dout : std_logic;
  signal den       : std_logic;
  signal dlast     : std_logic;
  signal clk_cnt   : natural range 0 to 9;
  signal data_cnt  : natural range 0 to 15;
begin
  --  Bit IO.
  --  Use 100Mhz/8 so 12.5Mhz.
  --  Start to transmit a bit when DEN = '1', signal last slot with DLAST.
  process (clk_100_i) is
  begin
    if rising_edge (clk_100_i) then
      if rst_n_i = '0' then
        clk_cnt <= 0;
        adc_sclk_o <= '0';
        dlast <= '0';
      else
        dlast <= '0';
        if clk_cnt = 0 then
          if den = '1' then
            --  Falling edge.
            adc_sclk_o <= '0';
            --  Start a bit
            adc_sdata_o <= dout;
            clk_cnt <= 1;
          else
            adc_sclk_o <= '1';
          end if;
        else
          if clk_cnt = 5 then
            --  Rising edge.
            adc_sclk_o <= '1';
            din <= adc_sdout_i;
          end if;
          if clk_cnt = 8 then
            --  Next bit is last bit
            dlast <= '1';
          end if;
          if clk_cnt = 9 then
            clk_cnt <= 0;
          else
            clk_cnt <= clk_cnt + 1;
          end if;
        end if;
      end if;
    end if;
  end process;

  process (clk_100_i) is
  begin
    if rising_edge (clk_100_i) then
      if rst_n_i = '0' then
        state <= IDLE;
        adc_cs_n_o <= '1';
        busy_o <= '0';
      else
        case state is
          when IDLE =>
            if start_i = '1' then
              --  Start tx.
              dout <= addr_i (7);
              adc_cs_n_o <= '0';
              data_cnt <= 7;
              den <= '1';
              busy_o <= '1';
              state <= ADDR;
            else
              adc_cs_n_o <= '1';
            end if;
          when ADDR =>
            if dlast = '1' then
              --  Next bit
              if data_cnt = 0 then
                state <= DATA;
                data_cnt <= 15;
                dout <= data_i (15);
              else
                dout <= addr_i (data_cnt - 1);
                data_cnt <= data_cnt - 1;
              end if;
            end if;
          when DATA =>
            if dlast = '1' then
              data_o (data_cnt) <= din;
              if data_cnt = 0 then
                --  End
                dout <= '0';
                den <= '0';
                data_cnt <= 7;
                state <= EXTRA;
              else
                dout <= data_i (data_cnt - 1);
                data_cnt <= data_cnt - 1;
              end if;
            end if;
          when EXTRA =>
            --  Extra idle cycle, with cs=1
            --  This adds a pause between two requests
            adc_cs_n_o <= '1';
            if data_cnt = 0 then
              busy_o <= '0';
              state <= IDLE;
            else
              data_cnt <= data_cnt - 1;
            end if;
        end case;
      end if;
    end if;
  end process;
end arch;
