--  Top level

library ieee;
use ieee.std_logic_1164.all;

library unisim;
use unisim.vcomponents.all;

use work.fsi_adc_pkg.all;

entity fsi_adc_top is
  port (
    --  Reset from backplane and powergood.
    fpga_reset_n_i : std_logic;

    --  Clock (100Mhz)
    fpga_clk_p_i : std_logic;
    fpga_clk_n_i : std_logic;

    --  Backplane IOs (lvds)
    --  Outputs (clock on lane 12, data on lanes 11-0)
    bp_io_p_o : out std_logic_vector(12 downto 0);
    bp_io_n_o : out std_logic_vector(12 downto 0);
    --  Inputs (clock on lane 1, data on lane 0)
    bp_io_p_i : std_logic_vector(1 downto 0);
    bp_io_n_i : std_logic_vector(1 downto 0);

    --  ADC data
    adc_outa_p_i : std_logic_vector(7 downto 0);
    adc_outa_n_i : std_logic_vector(7 downto 0);
    adc_outb_p_i : std_logic_vector(7 downto 0);
    adc_outb_n_i : std_logic_vector(7 downto 0);

    --  ADC bit clock
    adc_lclk_p_i : std_logic;
    adc_lclk_n_i : std_logic;

    --  ADC frame
    adc_adclk_p_i : std_logic;
    adc_adclk_n_i : std_logic;

    --  ADC control
    adc_reset_o : out std_logic;
    adc_pd_o : out std_logic;
    adc_sync_o : out std_logic;

    --  ADC SPI
    adc_sclk_o : out std_logic;
    adc_sdata_o : out std_logic;
    adc_sdout_i : in std_logic;
    adc_cs_n_o : out std_logic;

    --  TODO:
    --  fpga_pudc
    --  qspi_cs
    
    --  Trigger input
    trigger_p_i : std_logic;
    trigger_n_i : std_logic;

    --  I2C from backplane
    bp_sda_b : inout std_logic;
    bp_scl_i : in    std_logic;

    --  Analog front-end control
    afe_tia_ctrl_o : out std_logic_vector(8 downto 1);

    led0_o : out std_logic; --  The red led (on when '1')
    led1_o : out std_logic  --  The green led
    );
end fsi_adc_top;

architecture arch of fsi_adc_top is
  signal clk_100m : std_logic;
  signal clk_400m : std_logic;
  signal rst_n : std_logic;
  signal adc_locked : std_logic;
  signal adc_data : t_slv_8x12;
  signal sys_data_out : t_slv_8x12;

  signal sys_data_in : std_logic_vector(7 downto 0);
  signal sys_locked : std_logic;
  signal sys_ready : std_logic;
  signal sys_sampling : std_logic;

  signal spi_addr : std_logic_vector(7 downto 0);
  signal spi_data_in, spi_data_out : std_logic_vector(15 downto 0);
  signal spi_start, spi_busy : std_logic;

  signal trigger_in : std_logic;

  signal ctrl_dbg : std_logic_vector(3 downto 0);
  signal adc_dbg_slip : std_logic;
  signal adc_dbg_f : std_logic_vector(5 downto 0);
  signal adc_dbg_0a : std_logic_vector(5 downto 0);
  signal adc_dbg_0b : std_logic_vector(5 downto 0);
  signal rx_dbg : std_logic_vector(1 downto 0);
  signal dbg_clk : std_logic;

  signal adc_sclk :  std_logic;
  signal adc_sdata : std_logic;
  signal adc_cs_n :  std_logic;

begin
  --  Clocks
  inst_clk: entity work.fsi_adc_clk
    port map (
      fpga_reset_n_i => fpga_reset_n_i,
      fpga_clk_p_i => fpga_clk_p_i,
      fpga_clk_n_i => fpga_clk_n_i,
      clk_100_o => clk_100m,
      clk_400_o => clk_400m,
      rst_n_o => rst_n
    );

  --  ADC rx
  inst_adc_rx: entity work.fsi_adc_rx
    port map (
      clk_100_i => clk_100m,
      rst_n_i => rst_n,
      adc_outa_p_i => adc_outa_p_i,
      adc_outa_n_i => adc_outa_n_i,
      adc_outb_p_i => adc_outb_p_i,
      adc_outb_n_i => adc_outb_n_i,
      adc_lclk_p_i => adc_lclk_p_i,
      adc_lclk_n_i => adc_lclk_n_i,
      adc_adclk_p_i => adc_adclk_p_i,
      adc_adclk_n_i => adc_adclk_n_i,
      locked_o => adc_locked,
      data_o => adc_data,
      dbg_f => adc_dbg_f,
      dbg_0a => adc_dbg_0a,
      dbg_0b => adc_dbg_0b,
      dbg_slip => adc_dbg_slip,
      dbg_clk => dbg_clk
    );
  
  -- ADC control
  inst_adc_spi: entity work.fsi_adc_spi
    port map (
      clk_100_i => clk_100m,
      rst_n_i => rst_n,
      addr_i => spi_addr,
      data_i => spi_data_in,
      data_o => spi_data_out,
      start_i => spi_start,
      busy_o => spi_busy,
      adc_sclk_o => adc_sclk,
      adc_sdata_o => adc_sdata,
      adc_sdout_i => adc_sdout_i,
      adc_cs_n_o => adc_cs_n
    );

  adc_sclk_o <= adc_sclk;
  adc_sdata_o <= adc_sdata;
  adc_cs_n_o <= adc_cs_n;

  adc_pd_o <= '0';
  adc_sync_o <= '0';

  -- System tx
  inst_sys_tx: entity work.fsi_sys_tx
    port map (
      rst_n_i => rst_n,
      clk_100m_i => clk_100m,
      clk_400m_i => clk_400m,
      bp_io_p_o => bp_io_p_o,
      bp_io_n_o => bp_io_n_o,
      data_i => sys_data_out
    );

  -- System rx
  inst_sys_rx: entity work.fsi_sys_rx
    port map (
      clk_100_i => clk_100m,
      clk_400_i => clk_400m,
      rst_n_i => rst_n,
      bp_io_p_i => bp_io_p_i,
      bp_io_n_i => bp_io_n_i,
      locked_o => sys_locked,
      data_o => sys_data_in,
      dbg    => rx_dbg
    );

  --  Trigger
  inst_ibuf_trigger: IBUFDS
    port map (
      O => trigger_in,
      I => trigger_p_i,
      IB => trigger_n_i
  );

  --  State machine
  inst_ctrl: entity work.fsi_control
    port map (
      clk_100m_i => clk_100m,
      rst_n_i => rst_n,
      adc_reset_o => adc_reset_o,
      adc_locked_i => adc_locked,
      adc_spi_addr_o => spi_addr,
      adc_spi_data_o => spi_data_in,
      adc_spi_data_i => spi_data_out,
      adc_spi_start_o => spi_start,
      adc_spi_busy_i => spi_busy,
      afe_tia_ctrl_o => afe_tia_ctrl_o,
      sys_locked_i => sys_locked,
      sys_data_i => sys_data_in,
      sys_data_o => sys_data_out,
      sys_ready_o => sys_ready,
      sys_sampling_o => sys_sampling,
      adc_data_i => adc_data,
      trigger_i => trigger_in,
      dbg => ctrl_dbg
    );

  --led0_o <= fpga_reset_n_i;
  process (clk_100m)
    variable cnt : natural;
  begin
    if rising_edge(clk_100m) then
      if rst_n = '0' then
        led0_o <= '0';
        cnt := 0;
      else
        if sys_sampling = '1' then
          cnt := 10_000_000;
        elsif cnt /= 0 then
          cnt := cnt - 1;
        end if;

        --  Led is off while sampling,
        --  led is on if adc_locked.
        if adc_locked = '0' or cnt /= 0 then
          led0_o <= '1';
        else
          led0_o <= '0';
        end if;
      end if;
    end if;
  end process;

  --  Blinking led.
  gen_green_led: if true generate
    process (clk_100m) is
      variable cnt : natural;
    begin
      if rising_edge(clk_100m) then
        if rst_n = '0' then
          cnt := 0;
          led1_o <= '1';
        else
          cnt := cnt + 1;
          
          if (cnt = 10_000_000 - 1 and sys_ready = '0')
            or cnt = 50_000_000 - 1
          then
            led1_o <= '0';
          elsif cnt = 100_000_000 - 1 then
            led1_o <= '1';
            cnt := 0;
          end if;
        end if;
      end if;
    end process;
  end generate;

  --  TODO: flash program
  gen_ila: if False generate
    component ila_0
      port (
        clk    : in STD_LOGIC;
        probe0 : in STD_LOGIC_VECTOR(31 downto 0)
      );
    end component  ;
  begin
    inst_ila: ila_0
      port map (
        clk => dbg_clk, -- clk_100,
        probe0(0) => clk_100m, -- adc_locked,
        probe0(2 downto 1) => rx_dbg,
        probe0(6 downto 3) => ctrl_dbg,
        probe0(7) => adc_dbg_slip,
        probe0(13 downto 8) => adc_dbg_f,
        probe0(19 downto 14) => adc_dbg_0a,
        probe0(25 downto 20) => adc_dbg_0b,
        probe0(26) => spi_busy,
        probe0(27) => spi_start,
        probe0(28) => adc_sclk,
        probe0(29) => adc_cs_n,
        probe0(30) => adc_sdata,
        probe0(31) => adc_sdout_i
      );
  end generate;
end arch;
