library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.fsi_adc_pkg.all;

entity fsi_control is
  port (
    clk_100m_i : std_logic;
    rst_n_i : std_logic;

    adc_reset_o : out std_logic;
    adc_locked_i : std_logic;

    adc_spi_addr_o : out std_logic_vector(7 downto 0);
    adc_spi_data_o : out std_logic_vector(15 downto 0);
    adc_spi_data_i : std_logic_vector(15 downto 0);
    adc_spi_start_o : out std_logic;
    adc_spi_busy_i : in std_logic;

    afe_tia_ctrl_o : out std_logic_vector(8 downto 1);

    sys_locked_i : std_logic;
    sys_data_i : std_logic_vector(7 downto 0);
    sys_data_o : out t_slv_8x12;

    sys_ready_o : out std_logic;
    sys_sampling_o : out std_logic;

    adc_data_i : t_slv_8x12;

    trigger_i : std_logic;
    dbg : out std_logic_vector(3 downto 0)
  );    
end fsi_control;

architecture arch of fsi_control is
  type t_state is (
    RESET_0, RESET_1,
    INIT_ADC_0, INIT_ADC_1,
    WAIT_SYNC, IDLE, SEND_ADC, STATE_PRBS,
    PUSH_0, PUSH_1, PUSH_2, PUSH_3, EXEC);
  signal state : t_state;
  signal count : natural range 0 to 127;

  signal sync_data, msg_data : t_slv_8x12;
  signal msg_cmd : std_logic_vector(7 downto 0);
  signal msg_val : std_logic_vector(31 downto 0);
  signal cmd_done : std_logic;

  signal prbs, prbs_d : t_slv_8_arr (11 downto 0);

  --  Directly send ADC data to the system board.
  type t_send is (SEND_SYNC, SEND_MSG, SEND_SAMPLES, SEND_PRBS);
  signal send_sys : t_send;

  signal ready : std_logic;
  signal status : std_logic_vector(11 downto 0);
  signal sync_cnt : unsigned(3 downto 0);

  --  Data for a command.
  signal cmd_data : std_logic_vector(31 downto 0);

  --  ADC config
  type t_spi_config is record
    addr : std_logic_vector(7 downto 0);
    data : std_logic_vector(15 downto 0);
  end record;

  type t_spi_config_arr is array (natural range <>) of t_spi_config;

  constant c_adc_config : t_spi_config_arr (0 to 5):= (
    (x"26", b"0000_0000_0000_0000"),   --  pattern
--    (x"25", b"0000_0000_0100_0010"),   --  enable pattern
--    (x"46", b"1000_0010_0000_1101"),  -- 2 wires, two complement, MSB first, 12 bit
    (x"46", b"1000_0010_0000_0001"),    -- 2 wires, 12 bit
--    (x"28", b"00000000_00000000"),     -- Byte-wise
    (x"01", b"00000000_00000001"),     -- Enable read-out
    (x"46", b"00000000_00000000"),     -- Read reg
    (x"50", b"00000000_00000000"),     -- Read map
    (x"01", b"00000000_00000000")      -- Disable read-out
  );
begin
  --  Data for the backplane
  status (11) <= trigger_i;
  status (10) <= adc_locked_i;
  status (9)  <= sys_locked_i;
  status (8)  <= ready;
  status (7)  <= adc_spi_busy_i;
  status (6)  <= cmd_done;
  status (5)  <= '0';
  status (4)  <= '0';
  status (3 downto 0) <= std_logic_vector (sync_cnt);
  sync_data <= (7 => (others => '1'),
                6 => (others => '0'),
                5 => (others => '1'),
                4 => (others => '0'),
                3 => (others => '0'),
                2 => (others => '1'),
                1 => (others => '0'),
                0 => status);

  msg_data <= (7 => (others => '0'),
               6 => (others => '1'),
               5 => "0000" & msg_cmd,
               4 => "1111" & msg_val(31 downto 24),
               3 => "0000" & msg_val(23 downto 16),
               2 => "1111" & msg_val(15 downto 8),
               1 => "0000" & msg_val(7 downto 0),
               0 => status);

  sys_ready_o <= ready;

  process (clk_100m_i) is
  begin
    if rising_edge (clk_100m_i) then
      case send_sys is
        when SEND_SAMPLES =>
          sys_data_o <= adc_data_i;
        when SEND_SYNC =>
          sys_data_o <= sync_data;
        when SEND_MSG =>
          sys_data_o <= msg_data;
        when SEND_PRBS =>
          for i in 11 downto 0 loop
            for j in 7 downto 0 loop
              sys_data_o(j)(i) <= prbs_d(i)(j);
            end loop;
          end loop;
      end case;
    end if;
  end process;
          
  process (clk_100m_i) is
  begin
    if rising_edge (clk_100m_i) then
      --  Default values.
      adc_spi_start_o <= '0';
      cmd_done <= '0';
      ready <= '1';
      send_sys <= SEND_SYNC;
      sys_sampling_o <= '0';

      if rst_n_i = '0' then
        state <= RESET_0;
        adc_reset_o <= '1';
        count <= 10;
        ready <= '0';
        afe_tia_ctrl_o <= (others => '0');
        sync_cnt <= (others => '0');
        dbg <= x"0";
      else
        sync_cnt <= sync_cnt + 1;
        case state is
          when RESET_0 =>
            --  Apply reset to ADC.
            adc_reset_o <= '1';
            ready <= '0';
            if count = 0 then
              state <= RESET_1;
              count <= 20;
            else
              count <= count - 1;
            end if;
            dbg <= x"1";
          when RESET_1 =>
            --  Deassert ADC reset and wait.
            adc_reset_o <= '0';
            ready <= '0';
            if count = 0 then
              state <= INIT_ADC_0;
              count <= c_adc_config'left;
            else
              count <= count - 1;
            end if;    
            dbg <= x"2";
          when INIT_ADC_0 =>
            --  Program ADC.
            ready <= '0';
            if adc_spi_busy_i = '0' then
              if count > c_adc_config'right then
                state <= WAIT_SYNC;
              else
                adc_spi_addr_o <= c_adc_config(count).addr;
                adc_spi_data_o <= c_adc_config(count).data;
                adc_spi_start_o <= '1';
                state <= INIT_ADC_1;
                count <= count + 1;
              end if;
            end if;
            dbg <= x"3";

          when INIT_ADC_1 =>
            adc_spi_start_o <= '0';
            state <= INIT_ADC_0;
            ready <= '0';
            dbg <= x"4";

          when WAIT_SYNC =>
            --  Wait until all the lines are synchronized
            ready <= '0';
            if adc_locked_i = '1' and sys_locked_i = '1'
              and sys_data_i = c_sys_idle
            then
              state <= IDLE;
            end if;
            dbg <= x"5";

          when IDLE =>
            case sys_data_i is
              when c_sys_syncing =>
                --  Unexpected: need to resync ?
                state <= WAIT_SYNC;
              when c_sys_idle =>
                --  Expected
                null;
              when c_sys_sample_on =>
                --  Send the message.
                msg_cmd <= c_cmd_samples;
                msg_val <= (others => '0');
                send_sys <= SEND_MSG;
                sync_cnt <= (others => '0');
                state <= SEND_ADC;
              when c_sys_push8 =>
                state <= PUSH_3;
              when c_sys_push32 =>
                state <= PUSH_0;
              when c_sys_exec =>
                state <= EXEC;
              when others =>
                --  Unexpected command
                state <= WAIT_SYNC;
            end case;
            dbg <= x"6";

          when EXEC =>
            case sys_data_i is
              when c_exec_getdata =>
                --  Send a message (reply).
                msg_cmd <= c_cmd_reply;
                msg_val <= cmd_data;
                send_sys <= SEND_MSG;
                cmd_done <= '1';
                state <= IDLE;
              when c_exec_set_afe =>
                afe_tia_ctrl_o <= cmd_data (7 downto 0);
                cmd_done <= '1';
                state <= IDLE;
              when c_exec_tx_adc =>
                if adc_spi_busy_i = '0' then
                  adc_spi_addr_o <= cmd_data (31 downto 24);
                  adc_spi_data_o <= cmd_data (15 downto 0);
                  adc_spi_start_o <= '1';
                  --  Commands is done, but to wait while the spi is busy.
                  cmd_done <= '1';
                  state <= IDLE;
                else
                  state <= WAIT_SYNC;
                end if;
              when c_exec_rx_adc =>
                --  Send a message (reply).
                msg_cmd <= c_cmd_reply;
                msg_val <= x"0000" & adc_spi_data_i;
                send_sys <= SEND_MSG;
                cmd_done <= '1';
                state <= IDLE;
              when c_exec_prbs =>
                --  Send PRBS
                msg_cmd <= c_cmd_prbs;
                msg_val <= (others => '0');
                send_sys <= SEND_PRBS;
                sync_cnt <= (others => '0');
                state <= STATE_PRBS;
              when others =>
                --  Bad command.
                state <= WAIT_SYNC;
            end case;
            dbg <= x"7";

          when PUSH_0 =>
            cmd_data <= cmd_data(23 downto 0) & sys_data_i;
            state <= PUSH_1;
            dbg <= x"8";
          when PUSH_1 =>
            cmd_data <= cmd_data(23 downto 0) & sys_data_i;
            state <= PUSH_2;
          when PUSH_2 =>
            cmd_data <= cmd_data(23 downto 0) & sys_data_i;
            state <= PUSH_3;
          when PUSH_3 =>
            cmd_data <= cmd_data(23 downto 0) & sys_data_i;
            cmd_done <= '1';
            state <= IDLE;

          when SEND_ADC =>
            --  Send the samples.
            send_sys <= SEND_SAMPLES;
            sys_sampling_o <= '1';
            case sys_data_i is
              when c_sys_syncing =>
                --  Unexpected: need to resync ?
                state <= WAIT_SYNC;
              when c_sys_idle =>
                --  Expected
                null;
              when c_sys_sample_off =>
                state <= IDLE;
              when others =>
                state <= WAIT_SYNC;
            end case;
            dbg <= x"9";

          when STATE_PRBS =>
            --  Send the samples.
            send_sys <= SEND_PRBS;
            case sys_data_i is
              when c_sys_syncing =>
                --  Unexpected: need to resync ?
                state <= WAIT_SYNC;
              when c_sys_idle =>
                --  Expected
                null;
              when c_sys_sample_off =>
                state <= IDLE;
              when others =>
                state <= WAIT_SYNC;
            end case;
            dbg <= x"9";
          end case;
      end if;
    end if;
  end process;

  -- One PRBS per lane.
  gen_prbs: for i in 0 to 11 generate
    inst_prbs: entity work.fsi_prbs_gen
      generic map (
        init => std_logic_vector(to_unsigned(i, 4)) & x"1"
      )
      port map (
        clk_i => clk_100m_i,
        rst_n_i => rst_n_i,
        out_o => prbs(i)
      );
  end generate;

  process (clk_100m_i) is
  begin
    if rising_edge (clk_100m_i) then
      prbs_d <= prbs;
    end if;
  end process;

end arch;
