library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library unisim;
use unisim.vcomponents.all;

entity fsi_sys_rx is
  port (
    clk_100_i :     std_logic;
    clk_400_i :     std_logic;
    rst_n_i   :     std_logic;

    --  BP data
    bp_io_p_i :     std_logic_vector(1 downto 0);
    bp_io_n_i :     std_logic_vector(1 downto 0);

    locked_o  : out std_logic;
    data_o    : out std_logic_vector(7 downto 0);
    dbg : out std_logic_vector(1 downto 0)
  );
end fsi_sys_rx;

architecture arch of fsi_sys_rx is
  signal dlyin, dlyout : std_logic_vector(0 to 1);

  signal lds : std_logic_vector (0 to 1);

  type t_cntvalue is array (natural range <>) of unsigned (4 downto 0);
  signal cntvalues : t_cntvalue (0 to 1);

  type t_data is array (natural range <>) of std_logic_vector (7 downto 0);
  signal dout                       : t_data (0 to 1);

  signal rst                        : std_logic;
  signal bitslip                    : std_logic;
  signal lock_cnt                   : unsigned(3 downto 0);
  signal unlock_cnt                 : unsigned(2 downto 0);
  signal bit_locked, byte_locked    : std_logic;

  signal clk_400_n : std_logic;
begin
  --  Input pad (data)
  inst_ibuf_outa: IBUFDS_DIFF_OUT
    port map (
      O => dlyin (0),
      OB => dlyin (1),
      I => bp_io_p_i(0),
      IB => bp_io_n_i(0)
    );

  --  ISERDES
  rst <= not rst_n_i;
  clk_400_n <= not clk_400_i;

  --  IDELAY
  gen_idelay: for i in 0 to 1 generate
    inst_IDELAYE2 : IDELAYE2
      generic map (
      CINVCTRL_SEL => "FALSE",          -- Enable dynamic clock inversion (FALSE, TRUE)
      DELAY_SRC => "IDATAIN",           -- Delay input (IDATAIN, DATAIN)
      HIGH_PERFORMANCE_MODE => "FALSE", -- Reduced jitter ("TRUE"), Reduced power ("FALSE")
      IDELAY_TYPE => "VAR_LOAD",        -- FIXED, VARIABLE, VAR_LOAD, VAR_LOAD_PIPE
      IDELAY_VALUE => 0,                -- Input delay tap setting (0-31)
      PIPE_SEL => "FALSE",              -- Select pipelined mode, FALSE, TRUE
      REFCLK_FREQUENCY => 200.0,        -- IDELAYCTRL clock input frequency in MHz (190.0-210.0, 290.0-310.0).
      SIGNAL_PATTERN => "DATA"          -- DATA, CLOCK input signal
    )
    port map (
      CNTVALUEOUT => open,              -- output: Counter value output
      DATAOUT => dlyout(i),             -- output: Delayed data output
      C => clk_100_i,                   -- input: Clock input
      CE => '0',                        -- input: Active high enable increment/decrement input
      CINVCTRL => '0',                  -- input: Dynamic clock inversion input
      CNTVALUEIN => std_logic_vector(cntvalues(i)),       -- input: Counter value input
      DATAIN => '0',                    -- input: Internal delay data input
      IDATAIN => dlyin(i),              -- input: Data input from the I/O
      INC => '0',                       -- input: Increment / Decrement tap delay input
      LD => lds(i),                     -- input: Load IDELAY_VALUE input
      LDPIPEEN => '0',                  -- input: Enable PIPELINE register to load data input
      REGRST => rst                     -- input: Active-high reset tap-delay input
    );

    inst_ISERDESE2 : ISERDESE2
      generic map (
        DATA_RATE => "DDR", -- DDR, SDR
        DATA_WIDTH => 8,    -- Parallel data width (2-8,10,14)
        DYN_CLKDIV_INV_EN => "FALSE", -- Enable DYNCLKDIVINVSEL inversion (FALSE, TRUE)
        DYN_CLK_INV_EN => "FALSE",    -- Enable DYNCLKINVSEL inversion (FALSE, TRUE)
        INIT_Q1 => '0',               -- INIT_Q1 - INIT_Q4: Initial value on the Q outputs (0/1)
        INIT_Q2 => '0',
        INIT_Q3 => '0',
        INIT_Q4 => '0',
        INTERFACE_TYPE => "NETWORKING", -- MEMORY, MEMORY_DDR3, MEMORY_QDR, NETWORKING, OVERSAMPLE
        IOBDELAY => "BOTH", -- NONE, BOTH, IBUF, IFD
        NUM_CE => 2,        -- Number of clock enables (1,2)
        OFB_USED => "FALSE",-- Select OFB path (FALSE, TRUE)
        SERDES_MODE => "MASTER",    -- MASTER, SLAVE

        SRVAL_Q1 => '0',    -- SRVAL_Q1 - SRVAL_Q4: Q output values when SR is used (0/1)
        SRVAL_Q2 => '0',
        SRVAL_Q3 => '0',
        SRVAL_Q4 => '0'
      )
      port map (
        O => open,  -- 1-bit output: Combinatorial output

        Q1 => dout(i)(7),   -- Q1 - Q8: 1-bit (each) output: Registered data outputs
        Q2 => dout(i)(6),
        Q3 => dout(i)(5),
        Q4 => dout(i)(4),
        Q5 => dout(i)(3),
        Q6 => dout(i)(2),
        Q7 => dout(i)(1),
        Q8 => dout(i)(0),

        SHIFTOUT1 => open,-- SHIFTOUT1, SHIFTOUT2: 1-bit (each) output: Data width expansion output ports
        SHIFTOUT2 => open,
        BITSLIP => bitslip,

        CE1 => '1',-- CE1, CE2: 1-bit (each) input: Data register clock enable inputs
        CE2 => '1',
        CLKDIVP => '0', -- 1-bit input: TBD

        CLK => clk_400_i,    -- 1-bit input: High-speed clock
        CLKB => clk_400_n,   -- 1-bit input: High-speed secondary clock
        CLKDIV => clk_100_i, -- 1-bit input: Divided clock
        OCLK => '0',         -- 1-bit input: High speed output clock used when INTERFACE_TYPE="MEMORY"
        DYNCLKDIVSEL => '0', -- 1-bit input: Dynamic CLKDIV inversion
        DYNCLKSEL => '0',    -- 1-bit input: Dynamic CLK/CLKB inversion
        D => '0',            -- 1-bit input: Data input
        DDLY => dlyout(i),   -- 1-bit input: Serial data from IDELAYE2
        OFB => '0',          -- 1-bit input: Data feedback from OSERDESE2
        OCLKB => '0',        -- 1-bit input: High speed negative edge output clock
        RST => rst,

        SHIFTIN1 => '0',-- SHIFTIN1, SHIFTIN2: 1-bit (each) input: Data width expansion input ports
        SHIFTIN2 => '0'
      );
  end generate;

  --  Deskew
  blk_deskew: block
    signal dly: unsigned(0 to 2);
    constant c_histlog : natural := 5;
    signal balance : unsigned (c_histlog downto 0);
    signal history : std_logic_vector (2**c_histlog - 1 downto 0);

    --  True in case of error in locking.  May require a reset.
    signal lock_err : std_logic;
  begin
    process (clk_100_i) is
    begin
      if rising_edge(clk_100_i) then
        if rst_n_i = '0' then
          cntvalues(0) <= to_unsigned(12, 5);
          dly <= (others => '1');
          lds (0) <= '0';
          history <= (others => '0');
          balance <= (others => '0');
          bit_locked <= '0';
          lock_err <= '0';
        else
          lds (0) <= '0';
          --  Modify the delay only every 8 cycles.
          if dly = 0 then
            lock_err <= '0';
            --  Try to keep i+1 around the transition point.
            if dout(0) /= x"00" and dout(0) /= x"ff" then
              if dout(0) = (not dout(1)) then
                if cntvalues(0) >= 20 then
                  lock_err <= '1';
                end if;
                cntvalues(0) <= cntvalues(0) + 1;
                if history (0) = '1' then
                  --  An outgoing '1' is replaced by a '0'.
                  balance <= balance - 1;
                end if;
                history <= '0' & history (history'left downto 1);
              else
                if cntvalues (0) <= 3 then
                  lock_err <= '1';
                end if;
                cntvalues(0) <= cntvalues(0) - 1;
                if history (0) = '0' then
                  --  An outgoing '0' is replaced by a '1'.
                  balance <= balance + 1;
                end if;
                history <= '1' & history (history'left downto 1);
              end if;
              lds (0) <= '1';
              dly <= (others => '1');
            end if;
          else
            dly <= dly - 1;
          end if;
          if balance >= 2**(c_histlog - 1) - 1
            and balance <= 2**(c_histlog - 1) + 1
          then
            bit_locked <= '1';
          else
            bit_locked <= '0';
          end if;
        end if;
      end if;
    end process;

    cntvalues(1) <= cntvalues(0) + 8;
    lds (1) <= lds (0);
  end block;

  --  Bitslip control.
  process (clk_100_i) is
  begin
    if rising_edge(clk_100_i) then
      if rst = '1' then
        --  Not locked
        lock_cnt  <= (others => '0');
        unlock_cnt  <= (others => '0');
        bitslip <= '0';
        byte_locked <= '0';
      else
        if dout(0)(7 downto 3) = "10100" then
          --  Pattern is correct
          unlock_cnt  <= (others => '0');
          --  If locked for enough cycles, set the 'locked' register.
          if lock_cnt = (lock_cnt'range => '1') then
            byte_locked <= '1';
          else
            lock_cnt <= lock_cnt + 1;
          end if;
        else
          --  Wrong pattern.  Use an hysterisis to unlock.
          if lock_cnt /= 0 then
            if lock_cnt < 12 then
              --  Unlock if more than 4 bytes with incorrect pattern.
              byte_locked <= '0';
            end if;
            lock_cnt <= lock_cnt - 1;
          else
            --  Slip bit once every 8 cycles.
            if unlock_cnt = 0 then
              bitslip <= '1';
            else
              bitslip <= '0';
            end if;
            unlock_cnt <= unlock_cnt + 1;
          end if;
        end if;
      end if;
    end if;
  end process;

  locked_o <= bit_locked and byte_locked;

  dbg(0) <= bit_locked;
  dbg(1) <= byte_locked;
  
  data_o <= dout(0);
end arch;
