--  Data serialization to the system module.

library ieee;
use ieee.std_logic_1164.all;

library unisim;
use unisim.vcomponents.all;

use work.fsi_adc_pkg.all;

entity fsi_sys_tx is
  port (
    rst_n_i : std_logic;

    --  Clock (100Mhz) and the high-speed DDR clock (phase aligned)
    clk_100m_i : std_logic;
    clk_400m_i : std_logic;

    --  Backplane IOs (lvds)
    bp_io_p_o : out std_logic_vector(12 downto 0);
    bp_io_n_o : out std_logic_vector(12 downto 0);

    --  Data
    data_i : t_slv_8x12
    );
end fsi_sys_tx;

architecture arch of fsi_sys_tx is
  type t_slv_8x15 is array (0 to 7) of std_logic_vector(12 downto 0);
  signal bp_out : std_logic_vector(12 downto 0);
  signal edata : t_slv_8x15;
  signal rst : std_logic;

  --  Data parity.
  signal pdata : std_logic_vector(7 downto 0);

  function parity(v : std_logic_vector) return std_logic
  is
    variable res : std_logic := '0';
  begin
    for i in v'range loop
      res := res xor v(i);
    end loop;
    return res;
  end parity;

begin
  --  Compute parity
  process (clk_100m_i)
    variable v : std_logic_vector(11 downto 0);
  begin
    if rising_edge(clk_100m_i) then
      for i in 0 to 7 loop
        for j in 0 to 11 loop
          v (j) := data_i(i)(j);
        end loop;
        pdata(i) <= parity (v);
      end loop;
    end if;
  end process;

  --  Convert data: add parity.
  gen_cvdata: for i in 0 to 7 generate
    edata (i) <= pdata(i) & data_i(i);
  end generate;

  rst <= not rst_n_i;

  --  Serdes + pad
  --  TODO: odelay ?
  gen_ser: for i in 12 downto 0 generate
    inst_OSERDESE2 : OSERDESE2
      generic map (
        DATA_RATE_OQ => "DDR",
        DATA_RATE_TQ => "SDR",
        DATA_WIDTH => 8, -- Parallel data width (2-8,10,14)
        INIT_OQ => '0',  -- Initial value of OQ output (1'b0,1'b1)
        INIT_TQ => '0',  -- Initial value of TQ output (1'b0,1'b1)
        SERDES_MODE => "MASTER", -- MASTER, SLAVE
        SRVAL_OQ => '0', -- OQ output value when SR is used (1'b0,1'b1)
        SRVAL_TQ => '0', -- TQ output value when SR is used (1'b0,1'b1)
        TBYTE_CTL => "FALSE", -- Enable tristate byte operation (FALSE, TRUE)
        TBYTE_SRC => "FALSE", -- Tristate byte source (FALSE, TRUE)
        TRISTATE_WIDTH => 1  -- 3-state converter width (1,4)
        )
      port map (
        OFB => open,       -- 1-bit output: Feedback path for data (odelay)
        OQ => bp_out(i),   -- 1-bit output: Data path output
        SHIFTOUT1 => open,
        SHIFTOUT2 => open,
        TBYTEOUT => open,
        TFB => open,
        TQ => open,
        CLK => clk_400m_i,
        CLKDIV => clk_100m_i,
        D1 => edata(0)(i), -- D1-D8: Parallel data inputs, D1 output first
        D2 => edata(1)(i),
        D3 => edata(2)(i),
        D4 => edata(3)(i),
        D5 => edata(4)(i),
        D6 => edata(5)(i),
        D7 => edata(6)(i),
        D8 => edata(7)(i),
        OCE => '1', -- 1-bit input: Output data clock enable
        RST => rst,
        SHIFTIN1 => '0',
        SHIFTIN2 => '0',
        T1 => '0', -- T1 - T4: 1-bit (each) input: Parallel 3-state inputs
        T2 => '0',
        T3 => '0',
        T4 => '0',
        TBYTEIN => '0', -- 1-bit input: Byte group tristate
        TCE => '1'          -- 1-bit input: 3-state clock enable
        );
    
    inst_OBUFDS : OBUFDS
      generic map (
        IOSTANDARD => "DEFAULT", -- Specify the output
        SLEW => "FAST"
      )
      port map (
        O => bp_io_p_o(i),  -- Diff_p output
        OB => bp_io_n_o(i), -- Diff_n output
        I => bp_out(i)
      );
  end generate;
end arch;
