library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library unisim;
use unisim.vcomponents.all;

library xil_defaultlib;
use xil_defaultlib.fsi_adc_pkg.all;

entity sys_fsi_core is
  generic (
    g_NUM_BOARDS : natural := 8
  );
  port (
    clk_100m_i : std_logic;
    clk_400m_i : std_logic;
    clk_100m_oserdes_i : std_logic_vector(g_NUM_BOARDS downto 1);
    clk_400m_oserdes_i : std_logic_vector(g_NUM_BOARDS downto 1);
    rst_n_i : std_logic;
    io_rst_i : std_logic;

    --  Interfaces with FSI boards
    bp_io_p_i : t_slv_13_arr (g_NUM_BOARDS downto 1);
    bp_io_n_i : t_slv_13_arr (g_NUM_BOARDS downto 1);
    bp_io_p_o : out t_slv_2_arr (g_NUM_BOARDS downto 1);
    bp_io_n_o : out t_slv_2_arr (g_NUM_BOARDS downto 1);

    --  Force resync.
    resync_i : std_logic;

    --  Peripheral board enabled.
    board_en_i : std_logic_vector (g_NUM_BOARDS downto 1);

    --  Number of samples.
    samples_length_i : std_logic_vector(31 downto 0);

    --  Set to trigger an acquisition.
    trigger_i : std_logic;

    --  Switch to PRBS test mode.
    prbs_en_i : std_logic;

    --  Trigger status (not valid during acquisition).
    triggers_o : out std_logic_vector(g_NUM_BOARDS downto 1);

    --  Samples
    samples_o : out t_slv_8x12_arr (1 to g_NUM_BOARDS);
    samples_valid_o : out std_logic;

    --  System is up and ready (synchronized)
    all_ready_o : out std_logic;

    freeze_delays_i : std_logic;

    --  Status
    adc_locked_o : out std_logic_vector (g_NUM_BOARDS downto 1);
    sys_locked_o : out std_logic_vector (g_NUM_BOARDS downto 1);
    rx_locked_o : out std_logic_vector (g_NUM_BOARDS downto 1);
    fsi_ready_o : out std_logic_vector (g_NUM_BOARDS downto 1);
    spi_busy_o : out std_logic_vector (g_NUM_BOARDS downto 1);
    sampling_o : out std_logic;

    stat_sel_board_i : std_logic_vector(2 downto 0);
    stat_sel_ch_i : std_logic_vector(3 downto 0);
    stat_sel_dl_i : std_logic_vector(1 downto 0);
    stat_bit_locked_cnt_o : out std_logic_vector(31 downto 0);
    stat_byte_locked_cnt_o : out std_logic_vector(31 downto 0);
    stat_cntvalue_o : out std_logic_vector(31 downto 0);
    stat_prbs_o : out std_logic_vector(31 downto 0);
    stat_clock_o : out std_logic_vector (63 downto 0);
    stat_reset_i : in std_logic;

    delay_inc_i : std_logic;
    delay_dec_i : std_logic;
    delay_vtc_i : std_logic;

    --  Command or value to be sent.
    cmd_value_i : std_logic_vector (31 downto 0);
    --  Selected boards for the command or value.
    cmd_board_i : std_logic_vector (g_NUM_BOARDS downto 1);

    --  Pulse to send value or command.
    cmd_send_i : std_logic;
    cmd_exec_i : std_logic;
    cmd_ack_o : out std_logic;
    cmd_done_o : out std_logic_vector (g_NUM_BOARDS downto 1);
    rsp_value_o : out t_slv_32_arr (g_NUM_BOARDS downto 1)
  );
end sys_fsi_core;

architecture behav of sys_fsi_core is
  subtype t_boards is natural range g_NUM_BOARDS downto 1;
  signal samples: t_slv_8x12_arr (t_boards);
  signal lock : std_logic;
  signal locked : std_logic_vector (t_boards);
  signal dout : t_slv_8_arr (t_boards);

  signal trigger_d : std_logic;

  type t_slv_9_arr is array (natural range <>) of std_logic_vector(8 downto 0);
  
  type t_slv_32x13x8_arr is array (t_boards, 12 downto 0) of std_logic_vector(31 downto 0);

  signal stat_bit_locked_cnt, stat_byte_locked_cnt : t_slv_32x13x8_arr;
  signal stat_cntvalue : t_slv_9_arr (t_boards);
  signal stat_prbs_cnt : t_slv_32_arr(t_boards);
  signal stat_sel_board : t_boards;
  signal stat_clock : std_logic_vector(63 downto 0);
  signal stat_delay_vtc : std_logic;

  --  Status from fsi
  signal adc_locked, sys_locked, fsi_ready, spi_busy : std_logic_vector (t_boards);
  signal cmd_done_in, cmd_done : std_logic_vector (t_boards);
begin
  samples_o <= samples;
  cmd_done_o <= cmd_done;

  rx_locked_o <= locked;

  gen_boards: for i in t_boards generate
    signal stat_bit_locked, stat_byte_locked : std_logic_vector(12 downto 0);
    signal stat_delay_ce, stat_delay_inc : std_logic;
  begin
    inst_fsi_rx: entity work.sys_fsi_rx
      port map (
        clk_100m_i => clk_100m_i,
        clk_400m_i => clk_400m_i,
        rst_n_i => rst_n_i,
        io_rst_i => io_rst_i,
        bp_io_p_i => bp_io_p_i (i),
        bp_io_n_i => bp_io_n_i (i),
        dat_o => samples (i),
        locking_i => lock,
        locked_o => locked (i),
        resync_i => resync_i,
        freeze_delays_i => freeze_delays_i,
        stat_sel_ch_i => stat_sel_ch_i,
        stat_sel_dl_i => stat_sel_dl_i,
        stat_delay_ce_i => stat_delay_ce,
        stat_delay_inc_i => stat_delay_inc,
        stat_delay_vtc_i => stat_delay_vtc,
        stat_reset_i => stat_reset_i,
        stat_bit_locked_o => stat_bit_locked,
        stat_byte_locked_o => stat_byte_locked,
        stat_cntvalue_o => stat_cntvalue (i),
        stat_prbs_o => stat_prbs_cnt(i)
      );

    inst_fsi_tx: entity work.sys_fsi_tx
      port map (
        clk_100_i => clk_100m_oserdes_i(i),
        clk_400_i => clk_400m_oserdes_i(i),
        rst_n_i => rst_n_i,
        bp_io_p_o => bp_io_p_o (i),
        bp_io_n_o => bp_io_n_o (i),
        dat_i => dout (i)
      );

    --  Status bits
    triggers_o (i) <= samples (i)(0)(11);
    adc_locked (i) <= samples (i)(0)(10);
    sys_locked (i) <= samples (i)(0)(9);
    fsi_ready (i) <= samples (i)(0)(8);
    spi_busy (i) <= samples (i)(0)(7);
    cmd_done_in (i) <= samples (i)(0)(6);

    -- Stats
    process (clk_100m_i)
      function inc_saturate (v : std_logic_vector) return std_logic_vector is
      begin
        if v = (v'range => '1') then
          return v;
        else
          return std_logic_vector(unsigned(v) + 1);
        end if;
      end inc_saturate;
    begin
      if rising_edge (clk_100m_i) then
        if rst_n_i = '0' or stat_reset_i = '1' then
          for j in 0 to 12 loop
            stat_bit_locked_cnt (i, j) <= (others => '0');
            stat_byte_locked_cnt (i, j) <= (others => '0');
          end loop;
        else
          for j in 0 to 12 loop
            if stat_bit_locked(j) = '0' then
              stat_bit_locked_cnt (i, j) <= inc_saturate(stat_bit_locked_cnt (i, j));
            end if;
            if stat_byte_locked(j) = '0' then
              stat_byte_locked_cnt (i, j) <= inc_saturate(stat_byte_locked_cnt (i, j));
            end if;
          end loop;
        end if;
      end if;
    end process;

    process (clk_100m_i)
    begin
      if rising_edge (clk_100m_i) then
        stat_delay_ce <= '0';
        stat_delay_inc <= '0';
        if stat_sel_board = i and delay_vtc_i = '0' then
          --  Allow inc/dec only when vtc is off.
          if delay_inc_i = '1' and delay_dec_i = '0' then
            stat_delay_ce <= '1';
            stat_delay_inc <= '1';
          elsif delay_inc_i = '0' and delay_dec_i = '1' then
            stat_delay_ce <= '1';
            stat_delay_inc <= '0';
          end if;
        end if;
      end if;
    end process;
  end generate;

  stat_sel_board <= to_integer(unsigned(stat_sel_board_i) + 1);
  stat_bit_locked_cnt_o <= stat_bit_locked_cnt(stat_sel_board, to_integer(unsigned(stat_sel_ch_i)));
  stat_byte_locked_cnt_o <= stat_byte_locked_cnt(stat_sel_board, to_integer(unsigned(stat_sel_ch_i)));
  stat_cntvalue_o <= (31 downto 9 => '0') & stat_cntvalue(stat_sel_board);
  stat_prbs_o <= stat_prbs_cnt(stat_sel_board);
  
  process (clk_100m_i)
  begin
    if rising_edge (clk_100m_i) then
      if rst_n_i = '0' or stat_reset_i = '1' then
        stat_clock <= (others => '0');
      else
        stat_clock <= std_logic_vector(unsigned(stat_clock) + 1);
      end if;
    end if;
  end process;

  stat_clock_o <= stat_clock;

  process (clk_100m_i)
  begin
    if rising_edge (clk_100m_i) then
      stat_delay_vtc <= delay_vtc_i;
    end if;
  end process;

  blk_control: block
  is
    type t_state is (SYNCING, WAIT_TRIGGER, SAMPLING, POST_SAMPLING,
                    PUSH_0, PUSH_1, PUSH_2, PUSH_3,
                    PRBS_1, PRBS);
    signal state : t_state;
    signal count : unsigned (31 downto 0);
  begin
    process (clk_100m_i)
      variable sync : std_logic;
    begin
      if rising_edge (clk_100m_i) then
        cmd_ack_o <= '0';

        if rst_n_i = '0' then
          state <= SYNCING;
          all_ready_o <= '0';
          samples_valid_o <= '0';
          sampling_o <= '0';
          cmd_done <= (others => '0');
          dout <= (others => c_sys_syncing);
        else
          trigger_d <= trigger_i;

          case state is
            when SYNCING =>
              samples_valid_o <= '0';
              all_ready_o <= '0';
              trigger_d <= '0';
              --  Try to lock.
              lock <= '1';
              --  Check lock
              sync := '1';
              for i in t_boards loop
                if locked (i) = '1' then
                  --  RX is locked, so we can check status.
                  dout (i) <= c_sys_idle;
                  if adc_locked (i) = '0' or sys_locked (i) = '0' or fsi_ready (i) = '0' then
                    --  ADC not sync or sys not locked
                    if board_en_i (i) = '1' then
                      sync := '0';
                    end if;
                  end if;
                else
                  --  Still trying to sync.
                  dout (i) <= c_sys_syncing;
                  if board_en_i (i) = '1' then
                    sync := '0';
                  end if;
                end if;
              end loop;
              if sync = '1' then
                state <= WAIT_TRIGGER;
              end if;

            when WAIT_TRIGGER =>
              all_ready_o <= '1';
              lock <= '0';
              dout <= (others => c_sys_idle);
              if resync_i = '1'
                or (board_en_i and fsi_ready) /= board_en_i
              then
                state <= SYNCING;
              end if;

              --  Pass trigger to the peripheral boards.
              --  (only on the rising edge)
              dout <= (others => c_sys_idle);
              if trigger_i = '1' and trigger_d = '0' then
                --  Trigger.
                dout <= (others => c_sys_sample_on);
              elsif prbs_en_i = '1' then
                dout <= (others => c_sys_exec);
                state <= PRBS_1;
              elsif cmd_send_i = '1' then
                --  Send value.
                for i in dout'range loop
                  if cmd_board_i (i) = '1' then
                    dout (i) <= c_sys_push32;
                  end if;
                end loop;
                state <= PUSH_0;
                cmd_ack_o <= '1';
                cmd_done <= (others => '0');
              elsif cmd_exec_i = '1' then
                --  Execute command.
                for i in dout'range loop
                  if cmd_board_i (i) = '1' then
                    dout (i) <= c_sys_exec;
                  end if;
                end loop;
                state <= PUSH_3;
                cmd_ack_o <= '1';
                cmd_done <= (others => '0');
              else
                --  Check command from board 1
                --  TODO: check all in sync ?
                for i in t_boards loop
                  if board_en_i (i) = '1' then
                    --  Accumulate cmd_done answers.
                    cmd_done <= cmd_done or cmd_done_in;

                    --  Handle commands from FSI.
                    if samples (i)(7) = x"000" and samples (i)(6) = x"fff" then
                      if samples (i)(5)(7 downto 0) = c_cmd_samples then
                        --  Samples are coming...
                        count <= unsigned(samples_length_i);
                        samples_valid_o <= '1';
                        state <= SAMPLING;
                      elsif samples (i)(5)(7 downto 0) = c_cmd_reply then
                        rsp_value_o (i) <= samples (i)(4)(7 downto 0)
                                           & samples (i)(3)(7 downto 0)
                                           & samples (i)(2)(7 downto 0)
                                           & samples (i)(1)(7 downto 0);
                      end if;
                    end if;
                  end if;
                end loop;
              end if;

            when PUSH_0 =>
              dout <= (others => c_sys_idle);
              for i in dout'range loop
                if cmd_board_i (i) = '1' then
                  dout (i) <= cmd_value_i (31 downto 24);
                end if;
              end loop;
              state <= PUSH_1;
            when PUSH_1 =>
              dout <= (others => c_sys_idle);
              for i in dout'range loop
                if cmd_board_i (i) = '1' then
                  dout (i) <= cmd_value_i (23 downto 16);
                end if;
              end loop;
              state <= PUSH_2;
            when PUSH_2 =>
              dout <= (others => c_sys_idle);
              for i in dout'range loop
                if cmd_board_i (i) = '1' then
                  dout (i) <= cmd_value_i (15 downto 8);
                end if;
              end loop;
              state <= PUSH_3;
            when PUSH_3 =>
              dout <= (others => c_sys_idle);
              for i in dout'range loop
                if cmd_board_i (i) = '1' then
                  dout (i) <= cmd_value_i (7 downto 0);
                end if;
              end loop;
              state <= WAIT_TRIGGER;

            when PRBS_1 =>
              dout <= (others => c_exec_prbs);
              state <= PRBS;

            when PRBS =>
              if prbs_en_i = '0' then
                dout <= (others => c_sys_sample_off);
                state <= POST_SAMPLING;
                count <= to_unsigned (20, 32);
              else
                dout <= (others => c_sys_idle);
              end if;

            when SAMPLING =>
              if count = 1 then
                dout <= (others => c_sys_sample_off);
                state <= POST_SAMPLING;
                count <= to_unsigned (20, 32);
                samples_valid_o <= '0';
              else
                count <= count - 1;
              end if;

            when POST_SAMPLING =>
              --  Wait a little bit so that FSI boards don't send anymore data.
              dout <= (others => c_sys_idle);
              if count = 0 then
                state <= WAIT_TRIGGER;
              else
                count <= count - 1;
              end if;
          end case;
        end if;

        --  Update states.
        case state is
          when SYNCING
             | WAIT_TRIGGER
             | PUSH_0 to PUSH_3
             | PRBS_1
             | PRBS =>
            adc_locked_o <= adc_locked;
            sys_locked_o <= sys_locked;
            fsi_ready_o <= fsi_ready;
            spi_busy_o <= spi_busy;
            sampling_o <= '0';
          when SAMPLING
             | POST_SAMPLING =>
            sampling_o <= '1';
        end case;
      end if;
    end process;
  end block;
end behav;
