library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library unisim;
use unisim.vcomponents.all;

library xil_defaultlib;
use xil_defaultlib.fsi_adc_pkg.all;

entity sys_fsi_rx is
  port (
    clk_100m_i : std_logic;
    clk_400m_i : std_logic;
    rst_n_i : std_logic;
    io_rst_i : std_logic;

    --  Backplane IOs (lvds)
    bp_io_p_i : std_logic_vector(12 downto 0);
    bp_io_n_i : std_logic_vector(12 downto 0);

    --  Data out
    dat_o : out t_slv_8x12;

    --  Request to lock
    locking_i : std_logic;

    --  Locking status
    locked_o : out std_logic;

    --  Resync serdes
    resync_i : std_logic;

    --  Do not change delays when set.
    freeze_delays_i : std_logic;

    --  Statistics
    stat_sel_ch_i : std_logic_vector(3 downto 0);
    stat_sel_dl_i : std_logic_vector(1 downto 0);
    stat_delay_ce_i : std_logic;
    stat_delay_inc_i : std_logic;
    stat_delay_vtc_i : std_logic;
    stat_reset_i : std_logic;
    stat_cntvalue_o : out std_logic_vector(8 downto 0);
    stat_bit_locked_o : out std_logic_vector(12 downto 0);
    stat_byte_locked_o : out std_logic_vector(12 downto 0);
    stat_prbs_o : out std_logic_vector(31 downto 0)
  );
end sys_fsi_rx;

architecture arch of sys_fsi_rx is
  subtype t_serdes is natural range 12 downto 0;
  signal bp_io_in, bp_io_inb : std_logic_vector(t_serdes);

  signal dat_out : t_slv_8x12;
  signal parity : std_logic_vector(7 downto 0);

  signal id_cntvalueout, od_cntvalueout, is_cntvalueout, os_cntvalueout : t_slv_9_arr (t_serdes);

  signal clk_io_400, clk_io_100 : std_logic;

  signal byte_locked : std_logic_vector (t_serdes);
  signal bit_locked : std_logic_vector (t_serdes);

  signal locked : std_logic_vector(t_serdes);

  signal stat_prbs : t_slv_32_arr (t_serdes);
begin
  gen_inputs: for i in bp_io_p_i'range generate
    inst_ibufds: IBUFDS_DIFF_OUT
      port map (
        O  => bp_io_in (i),  -- 1-bit output: Buffer output
        OB => bp_io_inb (i),  -- 1-bit output: Buffer output
        I  => bp_io_p_i (i), -- 1-bit input: Diff_p buffer input
        IB => bp_io_n_i (i) -- 1-bit input: Diff_n buffer input
    );
  end generate;

  --  Not enough clocks per bank.
  gen_false: if false generate
    inst_bufg: BUFG
      port map (
        O => clk_io_400,
        I => bp_io_in (12)
      );

    inst_bufgce_div: BUFGCE_DIV
      generic map (
        BUFGCE_DIVIDE => 4,
        IS_CE_INVERTED => '0',  -- Optional inversion for CE
        IS_CLR_INVERTED => '0', -- Optional inversion for CLR
        IS_I_INVERTED => '0'    -- Optional inversion for I
      )
      port map (
        O => clk_io_100,    -- 1-bit output: Buffer
        CE => '1',          -- 1-bit input: Buffer enable
        CLR => '0',         -- 1-bit input: Asynchronous clear
        I => bp_io_in (12)  -- 1-bit input: Buffer
      );
  end generate;

  clk_io_400 <= clk_400m_i;
  clk_io_100 <= clk_100m_i;

  gen_idata: for i in t_serdes generate
    signal fifo_empty, fifo_rd_en : std_logic;
    signal fifo_empty2, fifo_rd_en2 : std_logic;
    signal bp_io_del, bp_io_del2 : std_logic;
    signal bp_io_out, bp_io_out_r, dout : std_logic_vector (7 downto 0);
    signal bp_io_out2 : std_logic_vector (7 downto 0);
    signal bp_io_out_concat: std_logic_vector (15 downto 0);
    signal load : std_logic;

    --  Number of IDELAY taps.  A tap is 2 to 15 ps.
    signal cntvalue : unsigned (8 downto 0);

    signal dly: unsigned(0 to 2);
    constant c_histlog : natural := 5;
    signal balance : unsigned (c_histlog downto 0);  -- Nbr of '1' in history
    signal history : std_logic_vector (2**c_histlog - 1 downto 0);
    signal casc_ret, casc_out : std_logic;
    signal casc_ret2, casc_out2 : std_logic;

    signal vtc : std_logic;

    constant c_lock_cnt_max : natural := 255;
    signal lock_cnt : natural range 0 to c_lock_cnt_max;
    signal bitslip_cnt : unsigned(2 downto 0);

    signal ce_data_odelay, ce_samp_odelay : std_logic;
    signal ce_delay : std_logic;
    signal prbs_err : std_logic_vector(7 downto 0);

    --  True in case of error in locking.  May require a reset.
    signal lock_err : std_logic;
  begin
    --  NOTE:  There are both an IDELAY and an ODELAY (cascaded mode).
    --  The IDELAY is used for setting the sample point, while the
    --  ODELAY is used in fixed mode for the offset between the sampling
    --  point and the probing point.
    --  As the IDELAY and ODELAY are chained, they must use the same
    --   DELAY_FORMAT, which is TIME so that the ODELAY can be set to
    --   a fixed and known delay.
    --  As TIME DELAY_FORMAT is used, VTC must be connected and not
    --   constant.  But because we are using VAR_LOAD on IDELAY, we
    --   don't really use the vtc adjustement so VTC is almost always 0.
    inst_data_idelay : IDELAYE3
      generic map (
        CASCADE => "MASTER",        -- Cascade setting
        DELAY_FORMAT => "TIME",   -- Units of the DELAY_VALUE (COUNT, TIME)
        DELAY_SRC => "IDATAIN",   -- Delay input IDATAIN -> IOB
        DELAY_TYPE => "VAR_LOAD", -- Set the type of tap delay line
        DELAY_VALUE => 0,
        IS_CLK_INVERTED => '0',
        IS_RST_INVERTED => '0',
        REFCLK_FREQUENCY => 400.0,
        SIM_DEVICE => "ULTRASCALE_PLUS", -- Set the device version for simu
        UPDATE_MODE => "ASYNC"
      )
      port map (
        CASC_OUT => casc_out,
        CNTVALUEOUT => id_cntvalueout (i),
        DATAOUT => bp_io_del,
        CASC_IN => '0',
        CASC_RETURN => casc_ret,
        CE => '0',
        CLK => clk_100m_i,
        CNTVALUEIN => std_logic_vector (cntvalue),
        DATAIN => '0',
        EN_VTC => vtc,
        IDATAIN => bp_io_in(i),
        INC => '0',
        LOAD => load,
        RST => io_rst_i
        );

    inst_data_odelay : ODELAYE3
      generic map (
        CASCADE => "SLAVE_END",   -- Cascade setting
        DELAY_FORMAT => "TIME",   -- Units of the DELAY_VALUE (COUNT, TIME)
        DELAY_TYPE => "VARIABLE", -- Set the type of tap delay line
        DELAY_VALUE => 100,
        IS_CLK_INVERTED => '0',
        IS_RST_INVERTED => '0',
        REFCLK_FREQUENCY => 400.0,
        SIM_DEVICE => "ULTRASCALE_PLUS", -- Set the device version for simu
        UPDATE_MODE => "ASYNC"
      )
      port map (
        CASC_OUT => open,
        CNTVALUEOUT => od_cntvalueout (i),
        DATAOUT => casc_ret,
        CASC_IN => casc_out,
        CASC_RETURN => '0',
        CE => ce_data_odelay,
        CLK => clk_100m_i,
        CNTVALUEIN => (others => '0'),
        ODATAIN => '0',
        EN_VTC => stat_delay_vtc_i,
        INC => stat_delay_inc_i,
        LOAD => '0',
        RST => io_rst_i
      );

    ce_delay <= '1' when stat_delay_ce_i = '1' and to_integer(unsigned(stat_sel_ch_i)) = i else '0';
    ce_data_odelay <= '1' when ce_delay = '1' and stat_sel_dl_i = "00" else '0';

    inst_data_iserdes: ISERDESE3
      generic map (
        DATA_WIDTH => 8,                -- Parallel data width (4,8)
        FIFO_ENABLE => "TRUE",          -- Enables the use of the FIFO
        FIFO_SYNC_MODE => "FALSE",      -- Always set to FALSE. TRUE is reserved for later use.
        IS_CLK_B_INVERTED => '1',       -- Optional inversion for CLK_B
        IS_CLK_INVERTED => '0',         -- Optional inversion for CLK
        IS_RST_INVERTED => '0',         -- Optional inversion for RST
        SIM_DEVICE => "ULTRASCALE_PLUS"
      )
      port map (
        FIFO_EMPTY => fifo_empty,       -- 1-bit output: FIFO empty flag
        INTERNAL_DIVCLK => open,        -- 1-bit output (do not connect)
        Q => bp_io_out,                 -- registered output
        CLK => clk_400m_i,              -- input: High-speed clock
        CLKDIV => clk_100m_i,           -- input: Divided Clock
        CLK_B => clk_400m_i,            -- input: Inversion of High-speed clock CLK
        D => bp_io_del,              -- input: Serial Data Input
        FIFO_RD_CLK => clk_100m_i,      -- input: FIFO read clock
        FIFO_RD_EN => fifo_rd_en,       -- input: Enables reading the FIFO when asserted
        RST => io_rst_i                  -- input: Asynchronous Reset
    );

    fifo_rd_en <= not fifo_empty;

    inst_samp_idelay : IDELAYE3
      generic map (
        CASCADE => "MASTER",      -- Cascade setting
        DELAY_FORMAT => "TIME",   -- Units of the DELAY_VALUE (COUNT, TIME)
        DELAY_SRC => "IDATAIN",   -- Delay input IDATAIN -> IOB
        DELAY_TYPE => "VAR_LOAD", -- Set the type of tap delay line
        DELAY_VALUE => 0,
        IS_CLK_INVERTED => '0',
        IS_RST_INVERTED => '0',
        REFCLK_FREQUENCY => 400.0,
        SIM_DEVICE => "ULTRASCALE_PLUS", -- Set the device version for simu
        UPDATE_MODE => "ASYNC"
      )
      port map (
        CASC_OUT => casc_out2,
        CNTVALUEOUT => is_cntvalueout (i),
        DATAOUT => bp_io_del2,
        CASC_IN => '0',
        CASC_RETURN => casc_ret2,
        CE => '0',
        CLK => clk_100m_i,
        CNTVALUEIN => std_logic_vector(cntvalue),
        IDATAIN => bp_io_inb (i),
        DATAIN => '0',
        EN_VTC => vtc,
        INC => '0',
        LOAD => load,
        RST => io_rst_i
      );

    inst_samp_odelay : ODELAYE3
      generic map (
        CASCADE => "SLAVE_END",   -- Cascade setting
        DELAY_FORMAT => "TIME",   -- Units of the DELAY_VALUE (COUNT, TIME)
        DELAY_TYPE => "VARIABLE", -- Set the type of tap delay line
        DELAY_VALUE => 400,       -- in ps; half is 625
        IS_CLK_INVERTED => '0',
        IS_RST_INVERTED => '0',
        REFCLK_FREQUENCY => 400.0,
        SIM_DEVICE => "ULTRASCALE_PLUS", -- Set the device version for simu
        UPDATE_MODE => "ASYNC"
      )
      port map (
        CASC_OUT => open,
        CNTVALUEOUT => os_cntvalueout (i),
        DATAOUT => casc_ret2,
        CASC_IN => casc_out2,
        CASC_RETURN => '0',
        CE => ce_samp_odelay,
        CLK => clk_100m_i,
        CNTVALUEIN => (others => '0'),
        ODATAIN => '0',
        EN_VTC => stat_delay_vtc_i,
        INC => stat_delay_inc_i,
        LOAD => '0',
        RST => io_rst_i
      );

    ce_samp_odelay <= '1' when ce_delay = '1' and stat_sel_dl_i = "10" else '0';

    inst_samp_iserdes: ISERDESE3
      generic map (
        DATA_WIDTH => 8,                -- Parallel data width (4,8)
        FIFO_ENABLE => "TRUE",          -- Enables the use of the FIFO
        FIFO_SYNC_MODE => "FALSE",      -- Always set to FALSE. TRUE is reserved for later use.
        IS_CLK_B_INVERTED => '1',       -- Optional inversion for CLK_B
        IS_CLK_INVERTED => '0',         -- Optional inversion for CLK
        IS_RST_INVERTED => '0',         -- Optional inversion for RST
        SIM_DEVICE => "ULTRASCALE_PLUS"
      )
      port map (
        FIFO_EMPTY => fifo_empty2,      -- 1-bit output: FIFO empty flag
        INTERNAL_DIVCLK => open,        -- 1-bit output (do not connect)
        Q => bp_io_out2,                -- registered output
        CLK => clk_400m_i,              -- input: High-speed clock
        CLKDIV => clk_100m_i,           -- input: Divided Clock
        CLK_B => clk_400m_i,            -- input: Inversion of High-speed clock CLK
        D => bp_io_del2,             -- input: Serial Data Input
        FIFO_RD_CLK => clk_100m_i,      -- input: FIFO read clock
        FIFO_RD_EN => fifo_rd_en2,      -- input: Enables reading the FIFO when asserted
        RST => io_rst_i                  -- input: Asynchronous Reset
    );

    fifo_rd_en2 <= not fifo_empty2;

    --  Deskewing
    process (clk_100m_i) is
    begin
      if rising_edge(clk_100m_i) then
        if rst_n_i = '0' or resync_i = '1' then
          cntvalue <= to_unsigned(250, 9);
          dly <= (others => '1');
          load <= '0';
          history <= (others => '0');
          balance <= (others => '0');
          bit_locked (i) <= '0';
          lock_err <= '0';
          vtc <= '1';
        else
          vtc <= '0';
          load <= '0';
          --  Modify the delay only every 8 cycles.
          if dly = 0 then
            lock_err <= '0';
            --  Try to keep i+1 around the transition point.
            if freeze_delays_i = '0' and bp_io_out /= x"00" and bp_io_out /= x"ff" then
              if (not bp_io_out) = bp_io_out2 then
                --  Increase delay
                if cntvalue >= 510 then
                  lock_err <= '1';
                end if;
                cntvalue <= cntvalue + 1;
                if history (0) = '1' then
                  --  Replacing a '1' by a '0', so one less '1'.
                  balance <= balance - 1;
                end if;
                history <= '0' & history (history'left downto 1);
              else
                --  Decrease delay
                if cntvalue <= 3 then
                  lock_err <= '1';
                end if;
                cntvalue <= cntvalue - 1;
                if history (0) = '0' then
                  --  Replacing a '0' by a '1', so one more '1'.
                  balance <= balance + 1;
                end if;
                history <= '1' & history (history'left downto 1);
              end if;
              --  Set the new cntvalue to the IDELAY
              load <= '1';
              --  Wait for 8 cycles before new adjustment.
              dly <= (others => '1');
            end if;
          else
            dly <= dly - 1;
          end if;
          if (balance <= 2**(c_histlog - 1) + 4)
            and (balance >= 2**(c_histlog - 1) - 4)
          then
            bit_locked (i) <= '1';
          else
            bit_locked (i) <= '0';
          end if;
        end if;
      end if;
    end process;

    process (clk_100m_i) is
    begin
      if rising_edge (clk_100m_i) then
        stat_bit_locked_o <= bit_locked;
        stat_byte_locked_o <= byte_locked;
      end if;
    end process;

    --  Bit slip
    bp_io_out_concat <= bp_io_out & bp_io_out_r;
    process (clk_100m_i) is
    begin
      if rising_edge (clk_100m_i) then
        dout <= std_logic_vector
                (shift_left (unsigned (bp_io_out_concat),
                              to_integer (bitslip_cnt))(15 downto 8));
        bp_io_out_r <= bp_io_out;
      end if;
    end process;

    --  Data lanes
    gen_nopar: if i < t_serdes'left generate
      gen_shuffle: for j in 7 downto 0 generate
        dat_out (j)(i) <= dout (j);
      end generate;
  
      byte_locked (i) <= '1' when dout (7 downto 1) = "1010010" else '0';
    end generate;

    --  Parity lane
    gen_par: if i = t_serdes'left generate
      byte_locked(i) <= '1' when dout = parity else '0';
    end generate;

    --  Bit slipping.
    process (clk_100m_i) is
    begin
      if rising_edge(clk_100m_i) then
        if rst_n_i = '0' or resync_i = '1' then
          locked (i) <= '0';
          lock_cnt <= 0;
          bitslip_cnt <= (others => '0');
        else
          if bit_locked (i) = '1'
            and byte_locked (i) = '1'
          then
            --  The pattern looks ok.  Wait for 16 cycles before changing the status.
            if lock_cnt = c_lock_cnt_max then
              locked (i) <= '1';
            else
              lock_cnt <= lock_cnt + 1;
            end if;
          else
            --  Not locked.
            locked (i) <= '0';
            if lock_cnt > 7 then
              --  Slip by 1 bit only if trying to lock.
              if locking_i = '1' then
                bitslip_cnt <= bitslip_cnt + 1;
              end if;
              lock_cnt <= 0;
            else
              --  Wait a little bit before locking.
              lock_cnt <= lock_cnt + 1;
            end if;
          end if;
        end if;
      end if;
    end process;

    --  PRBS receiver
    inst_prbs_chk: entity work.sys_prbs_chk
      port map (
        clk_i => clk_100m_i,
        rst_n_i => rst_n_i,
        dat_i => dout,
        err_o => prbs_err
      );

    --  PRBS counter
    process (clk_100m_i)
    begin
      if rising_edge(clk_100m_i) then
        if rst_n_i = '0' or stat_reset_i = '1' then
          stat_prbs (i) <= (others => '0');
        elsif prbs_err /= (7 downto 0 => '0')
          and stat_prbs (i) /= (31 downto 0 => '1')
        then
          stat_prbs (i) <= std_logic_vector
            (unsigned(stat_prbs (i)) + 1);
        end if;
      end if;
    end process;
  end generate;

  --  Parity
  process (clk_100m_i)
    variable p : std_logic;
  begin
    if rising_edge(clk_100m_i) then
      for i in 0 to 7 loop
        p := '0';
        for j in 0 to 11 loop
          p := p xor dat_out(i)(j);
        end loop;
        parity (i) <= p;
      end loop;
    end if;
  end process;

  locked_o <= '1' when locked = (12 downto 0 => '1') else '0';

  dat_o <= dat_out;

  --  i/o delay cnt value output.
  process (clk_100m_i) is
    variable idx : t_serdes;
  begin
    if rising_edge(clk_100m_i) then
      idx := to_integer(unsigned(stat_sel_ch_i));
      case stat_sel_dl_i is
        when "00" =>
          stat_cntvalue_o <= id_cntvalueout(idx);
        when "01" =>
          stat_cntvalue_o <= od_cntvalueout(idx);
        when "10" =>
          stat_cntvalue_o <= is_cntvalueout(idx);
        when "11" =>
          stat_cntvalue_o <= os_cntvalueout(idx);
        when others =>
      end case;
      stat_prbs_o <= stat_prbs (idx);
    end if;
  end process;
end arch;
