library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library xil_defaultlib;
use xil_defaultlib.fsi_adc_pkg.all;

entity sys_top_dma is
  port (
    --  Clock for the samples.
    clk_100m_i : std_logic;
    rst_100m_n_i : std_logic;

    --  Main clock (and reset)
    clk_300m_i : std_logic;
    rst_300m_n_i : std_logic;

    --  Incoming samples (use clk_100m_i)
    samples_i : t_slv_8x12_arr (0 to 3);
    samples_valid_i : std_logic;
    samples_start_i : std_logic;
  
    --  From the register maps (use clk_100m_i)
    dma_start_address_i : std_logic_vector(31 downto 0);
    dma_fifo_empty_o : out std_logic;
    dma_fifo_max_o : out std_logic_vector(7 downto 0);
    dma_fifo_errs_o : out std_logic_vector(7 downto 0);
    dma_axi_errs_o : out std_logic_vector(3 downto 0);

    --  For Data
    m_axi_dat_awid : out std_logic_vector(5 downto 0);
    m_axi_dat_awaddr : out std_logic_vector(48 downto 0);
    m_axi_dat_awlen : out std_logic_vector(7 downto 0);
    m_axi_dat_awsize : out std_logic_vector(2 downto 0);
    m_axi_dat_awburst : out std_logic_vector(1 downto 0);
    m_axi_dat_awprot : out std_logic_vector(2 downto 0);
    m_axi_dat_awcache : out std_logic_vector(3 downto 0);
    m_axi_dat_awuser : out std_logic_vector(3 downto 0);
    m_axi_dat_awvalid : out std_logic;
    m_axi_dat_awready : in std_logic;
    m_axi_dat_wdata : out std_logic_vector(127 downto 0);
    m_axi_dat_wstrb : out std_logic_vector(15 downto 0);
    m_axi_dat_wlast : out std_logic;
    m_axi_dat_wvalid : out std_logic;
    m_axi_dat_wready : in std_logic;
    m_axi_dat_bresp : in std_logic_vector(1 downto 0);
    m_axi_dat_bvalid : in std_logic;
    m_axi_dat_bready : out std_logic
  );
end sys_top_dma;

architecture arch of sys_top_dma is
  signal fifo_din, fifo_dout : std_logic_vector(383 + 1 downto 0);
  signal fifo_we, fifo_rd, fifo_full, fifo_empty : std_logic;
  signal fifo_cycle, fifo_cycle_next : natural range 0 to 2;
  signal data_out, data_out1, data_out2 : std_logic_vector(127 downto 0);
  signal data_out_next, data_out1_next, data_out2_next : std_logic_vector(127 downto 0);

  signal samples_start : std_logic;

  signal dma_running, dma_running_next: std_logic;
  signal dma_start, dma_done : std_logic;
  signal dma_start_burst : std_logic;

  signal m_axi_dat_awvalid_int  : std_logic;
  signal m_axi_dat_awvalid_next : std_logic;
  signal m_axi_dat_wvalid_int   : std_logic;
  signal m_axi_dat_wvalid_next  : std_logic;
  signal beat_acked              : std_logic;

  signal fifo_errs: unsigned(7 downto 0);
  signal fifo_max : unsigned(7 downto 0);
  signal wr_count : std_logic_vector(7 downto 0);

  --  Burst length is limited to 16 on the DDR part.
  constant c_BURST_BYTE : natural := 256;

  --  Target address for the next burst (in bytes).
  signal dma_address : unsigned (31 downto 0);

  signal samples_valid_d : std_logic;

  signal data_out_valid, data_out_valid_next : std_logic;

  signal axi_id : unsigned(5 downto 0);

  --  Number of data transfers for the current burst.
  signal dma_cur_length, dma_cur_length_next : unsigned (7 - 4 downto 0);
  constant c_DMA_LEN_LAST : unsigned (7 - 4 downto 0) := (others => '1');

  type t_state is (IDLE, PRE_START, BURST, DONE);
  signal state : t_state;
begin
  --  Synchronizer for samples_start_i (from 100m to 300m)
  inst_sync_start: entity work.gc_sync_ffs
    generic map (
      g_sync_edge => "positive"
    )
    port map (
      clk_i => clk_300m_i,
      rst_n_i => rst_300m_n_i,
      data_i => samples_start_i,
      synced_o => open,
      npulse_o => open,
      ppulse_o => samples_start
    );

  inst_fifo_dat0: entity work.generic_async_fifo
    generic map (
      g_data_width => 384 + 1,
      g_size => 256,
      g_show_ahead => True,
      g_with_rd_empty => True,
      g_with_rd_full => False,
      g_with_rd_almost_empty => False,
      g_with_rd_almost_full => False,
      g_with_rd_count => False,
      g_with_wr_empty => True,
      g_with_wr_full => True,
      g_with_wr_almost_empty => False,
      g_with_wr_almost_full => False,
      g_with_wr_count => True,
      g_almost_empty_threshold => 250,
      g_almost_full_threshold => 250
    )
    port map (
      rst_n_i => rst_100m_n_i,
      clk_wr_i => clk_100m_i,
      d_i => fifo_din,
      we_i => fifo_we,
      wr_empty_o => dma_fifo_empty_o,
      wr_full_o => fifo_full,
      wr_almost_empty_o => open,
      wr_almost_full_o => open,
      wr_count_o => wr_count,
      clk_rd_i => clk_300m_i,
      q_o => fifo_dout,
      rd_i => fifo_rd,
      rd_empty_o => fifo_empty,
      rd_full_o => open,
      rd_almost_empty_o => open,
      rd_almost_full_o => open,
      rd_count_o => open
    );

  --  FIFO input: samples + valid.
  --  Also update the status.
  process (clk_100m_i)
  begin
    if rising_edge(clk_100m_i) then
      if rst_100m_n_i = '0' then
        fifo_we <= '0';
        fifo_errs <= (others => '0');
        fifo_max <= (others => '0');
        samples_valid_d <= '0';
      else
        --  Write all the samples + one with valid set to 0.
        fifo_we <= samples_valid_i or samples_valid_d;
        samples_valid_d <= samples_valid_i;

        --  Fifo stats
        if samples_start_i = '1' then
          fifo_errs <= (others => '0');
          fifo_max <= (others => '0');
        else
          if fifo_full = '1' and samples_valid_i = '1' and fifo_errs /= x"ff" then
            fifo_errs <= fifo_errs + 1;
          end if;
          if unsigned(wr_count) > fifo_max then
            fifo_max <= unsigned(wr_count);
          end if;
        end if;
      end if;

      -- Linearize the data for the fifo.
      for i in 0 to 3 loop
        for j in 0 to 7 loop
          fifo_din(12*8*i + 12*j + 11 downto 12*8*i + 12*j) <= samples_i(i)(j);
        end loop;
      end loop;
      --  Valid bit.
      fifo_din(384) <= samples_valid_i;
    end if;
  end process;

  dma_fifo_max_o <= std_logic_vector(fifo_max);
  dma_fifo_errs_o <= std_logic_vector(fifo_errs);

  beat_acked <= m_axi_dat_wvalid_int and m_axi_dat_wready;

  dma_cur_length_next <= dma_cur_length + 1 when beat_acked = '1'
                         else dma_cur_length;

  --  FIFO out to AXI
  process (dma_running, fifo_empty, beat_acked, fifo_cycle, fifo_dout,
           m_axi_dat_wvalid_int, dma_done, dma_start,
           dma_cur_length, data_out, data_out1, data_out2, data_out_valid)
  begin
    fifo_rd <= '0';

    --  By default, keep values as they are.
    dma_running_next <= dma_running;
    m_axi_dat_wvalid_next <= m_axi_dat_wvalid_int;
    fifo_cycle_next <= fifo_cycle;
    data_out_next <= data_out;
    data_out1_next <= data_out1;
    data_out2_next <= data_out2;
    data_out_valid_next <= data_out_valid;

    --  Read data either when the first data is available (and therefore
    --  start DMA) or when the DMA is running and the previous data was
    --  sent on the bus.
    if dma_start = '1'
      or (dma_running ='1' and beat_acked = '1')
    then
      --  By default, data are sent.
      dma_running_next <= '1';
      m_axi_dat_wvalid_next <= '1';

      --  On the first cycle, read the fifo and scatter the data for
      --  3 cycles.
      if fifo_cycle = 0 then
        fifo_cycle_next <= 1;

        if data_out_valid = '1' and fifo_empty = '1' then
          --  Fifo is not ready but data have to be sent from the fifo.
          --  Just freeze.
          m_axi_dat_wvalid_next <= '0';
          fifo_cycle_next <= 0;
        elsif data_out_valid = '1' or dma_start = '1' then
          --  Fifo data are valid.

          --  fifo_dout length is 12*32
          for i in 0 to 15 loop
            data_out_next (8*i + 7 downto 8*i) <= fifo_dout(12*i + 11 downto 12*i + 4);
            data_out1_next(8*i + 7 downto 8*i) <= fifo_dout(192 + 12*i + 11 downto 192 + 12*i + 4);
            data_out2_next(4*i + 3 downto 4*i) <= fifo_dout(12*i + 3 downto 12*i);
            data_out2_next(64 + 4*i + 3 downto 64 + 4*i) <= fifo_dout(192 + 12*i + 3 downto 192 + 12*i);
          end loop;
          data_out_valid_next <= fifo_dout(384);
          --  Data are read.
          fifo_rd <= '1';
        else
          --  There are still data to be sent (to finish the burst), but
          --  not from the fifo.
          data_out_next <= (others => '0');
          data_out1_next <= (others => '0');
          data_out2_next <= (others => '0');
          data_out_valid_next <= '0';
        end if;
      else
        --  Second and third cycles
        data_out_next  <= data_out1;
        data_out1_next <= data_out2;
        data_out2_next <= (others => '1');
        if fifo_cycle = 1 then
          fifo_cycle_next <= 2;
        else
          fifo_cycle_next <= 0;
        end if;
      end if;
    end if;

    if (dma_cur_length = c_DMA_LEN_LAST)
      and beat_acked = '1' and dma_done = '1'
    then
      --  Last data transfer.
      dma_running_next <= '0';
      fifo_cycle_next <= 0;
      m_axi_dat_wvalid_next <= '0';
    end if;
  end process;

  process (clk_300m_i)
  begin
    if rising_edge(clk_300m_i) then
      if rst_300m_n_i = '0' then
        dma_running <= '0';
        fifo_cycle <= 0;
        data_out_valid <= '0';
      else
        dma_running <= dma_running_next;
        fifo_cycle <= fifo_cycle_next;
        data_out_valid <= data_out_valid_next;
      end if;
      data_out <= data_out_next;
      data_out1 <= data_out1_next;
      data_out2 <= data_out2_next;

      if dma_cur_length_next = c_DMA_LEN_LAST then
        --  Last data transfer in the burst.
        m_axi_dat_wlast <= '1';
      else
        m_axi_dat_wlast <= '0';
      end if;
    end if;
  end process;

  m_axi_dat_awvalid <= m_axi_dat_awvalid_int;
  m_axi_dat_wvalid <= m_axi_dat_wvalid_int;

  m_axi_dat_awid <= std_logic_vector(axi_id);
  m_axi_dat_awaddr <=
    (48 downto 32 => '0') & std_logic_vector (dma_address);
  m_axi_dat_awlen <= x"0f";
  m_axi_dat_awsize <= "100";  -- 16 bytes
  m_axi_dat_awburst <= "01";  -- INCR
  m_axi_dat_awprot <= "000";
  m_axi_dat_awcache <= "0011";
  m_axi_dat_awuser <= (others => '0');

  --  Send a command either when sampling starts or when the previous
  --  burst is done.
  m_axi_dat_awvalid_next <=
    dma_start_burst or (m_axi_dat_awvalid_int and not m_axi_dat_awready);

  m_axi_dat_wstrb <= (others => '1');
  m_axi_dat_wdata <= data_out;

  process (clk_300m_i)
  begin
    if rising_edge(clk_300m_i) then
      dma_start_burst <= '0';
      dma_start <= '0';

      if rst_300m_n_i = '0' then
        axi_id <= (others => '0');
        dma_done <= '0';
        state <= IDLE;
      else
        dma_cur_length <= dma_cur_length_next;
        dma_done <= '0';

        case state is
          when IDLE =>
            if samples_start = '1' then
              --  Samples will come soon.  Prepare DMA.
              --  Truncate the address so that no 4KB pages are crossed.
              dma_address <= unsigned (dma_start_address_i and x"ffff_ff00");
              dma_cur_length <= (others => '0');
              state <= PRE_START;
            end if;

          when PRE_START =>
            if fifo_empty = '0' then
              --  First sample arrived: start a burst.
              dma_start_burst <= '1';
              dma_start <= '1';
              state <= BURST;
            end if;

          when BURST =>
            if m_axi_dat_awvalid_int = '1' and m_axi_dat_awready = '1' then
              --  When AW has been acked, prepare the next transfer
              dma_address <= dma_address + c_BURST_BYTE;
              axi_id <= axi_id + 1;
            end if;

            if dma_cur_length = (c_DMA_LEN_LAST - 2)
              and beat_acked = '1'
            then
              --  Start a new burst if there are still samples.
              --  Note: at worst, there is one burst of extra data if
              --  samples_valid goes down on the next cycle.
              if data_out_valid = '1' then
                dma_start_burst <= '1';
              else
                dma_done <= '1';
                state <= DONE;
              end if;
            end if;

          when DONE =>
            --  Wait until the end of the DMA transfer.
            dma_done <= '1';
            if dma_running = '0' then
              state <= IDLE;
            end if;
        end case;
      end if;
    end if;
  end process;

  process (clk_300m_i)
  begin
    if rising_edge(clk_300m_i) then
      if rst_300m_n_i = '0' then
        m_axi_dat_awvalid_int <= '0';
        m_axi_dat_wvalid_int <= '0';
        m_axi_dat_bready <= '0';
      else
        --  Delayed version of valid signals.
        m_axi_dat_awvalid_int <= m_axi_dat_awvalid_next;
        m_axi_dat_wvalid_int <= m_axi_dat_wvalid_next;

        --  Always ready for the status.
        m_axi_dat_bready <= '1';

        --  TODO: capture bresp
      end if;
    end if;
  end process;
end arch;
