library ieee;
use ieee.std_logic_1164.all;

use work.fsi_adc_pkg.all;

entity periph_adc_tb is
end periph_adc_tb;

architecture arch of periph_adc_tb is
  signal fpga_reset_n : std_logic;
  signal clk_100m : std_logic;
  signal clk_400m : std_logic;
  signal clk_400m_int : std_logic := '0';
  signal io_rst : std_logic;

  signal fpga_clk_p, fpga_clk_n : std_logic;

  signal bp_io_p_out, bp_io_n_out : t_slv_13_arr (8 downto 1);
  signal bp_io_p_out_d, bp_io_n_out_d : t_slv_13_arr (8 downto 1);
  signal bp_io_p_in : t_slv_2_arr (8 downto 1);
  signal bp_io_n_in : t_slv_2_arr (8 downto 1);

  --  ADC data
  signal adc_outa_p : std_logic_vector(7 downto 0);
  signal adc_outa_n : std_logic_vector(7 downto 0);
  signal adc_outb_p : std_logic_vector(7 downto 0);
  signal adc_outb_n : std_logic_vector(7 downto 0);

  --  ADC bit clock
  signal adc_lclk_p   : std_logic;
  signal adc_lclk_n   : std_logic;

  --  ADC frame
  signal adc_adclk_p  : std_logic;
  signal adc_adclk_n  : std_logic;

  --  ADC control
  signal adc_reset    : std_logic;
  signal adc_pd       : std_logic;
  signal adc_sync     : std_logic;

  --  ADC SPI
  signal adc_sclk     : std_logic;
  signal adc_sdata    : std_logic;
  signal adc_sdout    : std_logic;
  signal adc_cs_n     : std_logic;

  --  TODO:
  --  fpga_pudc
  --  qspi_cs

  --  Trigger input
  signal trigger_p    : std_logic;
  signal trigger_n    : std_logic;

  --  I2C from backplane
  signal bp_sda       : std_logic;
  signal bp_scl       : std_logic;

  --  Analog front-end control
  signal afe_tia_ctrl : std_logic_vector(8 downto 1);

  signal samples : t_slv_8x12;
  signal samples_valid : std_logic;
  signal fsi_ready : std_logic;
  signal trigger, prbs_en : std_logic;

  signal cmd_value : std_logic_vector (31 downto 0);
  signal cmd_board, cmd_done : std_logic_vector (8 downto 1);

  --  Pulse to send value or command.
  signal cmd_send, cmd_exec : std_logic;
  signal cmd_ack : std_logic;

  signal rsp_value : t_slv_32_arr (8 downto 1);

begin
  clk_400m_int <= not clk_400m_int after 1250 ps;

  process (clk_400m_int)
    variable count : natural := 0;
  begin
    clk_400m <= clk_400m_int;
    if clk_400m_int = '1' then
      count := count + 1;
      if count = 4 then
        clk_100m <= '0';
        count := 0;
      elsif count = 2 then
        clk_100m <= '1';
      else
      end if;
    end if;
  end process;

  fpga_clk_p <= clk_100m;
  fpga_clk_n <= not clk_100m;

  fpga_reset_n <= '0', '1' after 40 ns;
  io_rst <= '1', '0' after 40 ns;

  trigger_n <= not trigger;
  trigger_p <= trigger;

  --  Commands and trigger generation
  process
    procedure wait_cmd_done is
    begin
      --  Wait until the command has been processed.
      loop
        wait until rising_edge(clk_100m);
        exit when (cmd_done and cmd_board) = cmd_board;
      end loop;
    end wait_cmd_done;

    procedure send_data (data : std_logic_vector(31 downto 0)) is
    begin
      cmd_board <= (1 => '1', others => '0');
      cmd_value <= data;
      cmd_send <= '1';
      wait until rising_edge(clk_100m);
      cmd_send <= '0';
      wait until rising_edge(clk_100m);
      assert cmd_ack = '1' report "send_data cmd_ack";
      assert cmd_done = (8 downto 1 => '0') report "send_data cmd done";
      wait_cmd_done;
    end send_data;

    procedure exec_cmd (cmd : std_logic_vector(7 downto 0)) is
    begin
      cmd_board <= (1 => '1', others => '0');
      cmd_value <= x"000000" & cmd;
      cmd_exec <= '1';
      wait until rising_edge(clk_100m);
      cmd_exec <= '0';
      wait until rising_edge(clk_100m);
      assert cmd_ack = '1' report "exec_cmd cmd_ack" severity note;
      assert cmd_done = (8 downto 1 => '0') report "exec_cmd cmd_done";
      wait_cmd_done;
    end exec_cmd;
  begin
    report "Start testbench";

    trigger <= '0';
    cmd_exec <= '0';
    cmd_send <= '0';
    prbs_en <= '0';
    wait until rising_edge(clk_100m) and fsi_ready = '1';
    for i in 1 to 25 loop
      wait until rising_edge(clk_100m);
    end loop;

    report "Start commands";

    --  Send spi
    send_data (x"ff_00_1234");
    exec_cmd (c_exec_tx_adc);
    report "spi tx done";

    --  Execute set afe command
    send_data (x"ef_cd_ab_ff");
    exec_cmd (c_exec_set_afe);
    report "set_afe done";

    exec_cmd (c_exec_rx_adc);
    assert rsp_value (1)(15 downto 0) = x"de15" report "rsp value";
    report "spi rx done";

    trigger <= '1';
    wait until rising_edge(clk_100m);
    trigger <= '0';
    report "triggered";

    wait;
  end process;

  inst_fsi: entity work.fsi_adc_top
    port map (
      fpga_reset_n_i => fpga_reset_n,
      fpga_clk_p_i => fpga_clk_p,
      fpga_clk_n_i => fpga_clk_n,
      bp_io_p_o => bp_io_p_out(1),
      bp_io_n_o => bp_io_n_out(1),
      bp_io_p_i => bp_io_p_in(1),
      bp_io_n_i => bp_io_n_in(1),
      adc_outa_p_i => adc_outa_p,
      adc_outa_n_i => adc_outa_n,
      adc_outb_p_i => adc_outb_p,
      adc_outb_n_i => adc_outb_n,
      adc_lclk_p_i => adc_lclk_p,
      adc_lclk_n_i => adc_lclk_n,
      adc_adclk_p_i => adc_adclk_p,
      adc_adclk_n_i => adc_adclk_n,
      adc_reset_o => adc_reset,
      adc_pd_o => adc_pd,
      adc_sync_o => adc_sync,
      adc_sclk_o => adc_sclk,
      adc_sdata_o => adc_sdata,
      adc_sdout_i => adc_sdout,
      adc_cs_n_o => adc_cs_n,
      trigger_p_i => trigger_p,
      trigger_n_i => trigger_n,
      bp_sda_b => bp_sda,
      bp_scl_i => bp_scl,
      afe_tia_ctrl_o => afe_tia_ctrl
      );

  inst_ads5295: entity work.ads5295
    port map (
      clk_p_i => fpga_clk_p,
      clk_n_i => fpga_clk_n,
      adc_outa_p_o => adc_outa_p,
      adc_outa_n_o => adc_outa_n,
      adc_outb_p_o => adc_outb_p,
      adc_outb_n_o => adc_outb_n,
      adc_lclk_p_o => adc_lclk_p,
      adc_lclk_n_o => adc_lclk_n,
      adc_adclk_p_o => adc_adclk_p,
      adc_adclk_n_o => adc_adclk_n,
      adc_reset_i => adc_reset,
      adc_pd_i => adc_pd,
      adc_sync_i => adc_sync,
      adc_sclk_i => adc_sclk,
      adc_sdata_i => adc_sdata,
      adc_sdout_o => adc_sdout,
      adc_cs_n_i => adc_cs_n
    );

  bp_io_n_out_d <= bp_io_n_out after 500 ps;
  bp_io_p_out_d <= bp_io_p_out after 500 ps;

  inst_sys_core: entity work.sys_fsi_core
    port map (
      clk_100m_i => clk_100m,
      clk_400m_i => clk_400m,
      rst_n_i => fpga_reset_n,
      io_rst_i => io_rst,
      bp_io_p_i => bp_io_p_out_d,
      bp_io_n_i => bp_io_n_out_d,
      bp_io_p_o => bp_io_p_in,
      bp_io_n_o => bp_io_n_in,
      resync_i => '0',
      board_en_i => (1 => '1', others => '0'),
      samples_length_i => x"0000_00c8",
      trigger_i => trigger,
      prbs_en_i => prbs_en,
      samples_o (1) => samples,
      samples_valid_o => samples_valid,
      all_ready_o => fsi_ready,
      adc_locked_o => open,
      sys_locked_o => open,
      rx_locked_o => open,
      fsi_ready_o => open,
      freeze_delays_i => '0',
      stat_sel_board_i => b"000",
      stat_sel_ch_i => x"0",
      stat_sel_dl_i => "00",
      stat_reset_i => '0',
      delay_inc_i => '0',
      delay_dec_i => '0',
      delay_vtc_i => '0',
      cmd_value_i => cmd_value,
      cmd_board_i => cmd_board,
      cmd_send_i => cmd_send,
      cmd_exec_i => cmd_exec,
      cmd_ack_o => cmd_ack,
      cmd_done_o => cmd_done,
      rsp_value_o => rsp_value
    );
end arch;
