/*
 * helloworld.c: simple test application
 *
 * This application configures UART 16550 to baud rate 9600.
 * PS7 UART (Zynq) is not initialized by this application, since
 * bootrom/bsp configures it to baud rate 115200
 *
 * ------------------------------------------------
 * | UART TYPE   BAUD RATE                        |
 * ------------------------------------------------
 *   uartns550   9600
 *   uartlite    Configurable only in HW design
 *   ps7_uart    115200 (configured by bootrom/bsp)
 */

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include "xil_cache.h"
#include "xgpiops.h"
#include "xparameters.h"
#include <xil_types.h>
#include <xiicps.h>
#include <xiicps_hw.h>
#include <xuartps_hw.h>

#include "sys_top_sub_regs.h"

#undef GPIO_MASK_DATA_5_MSW_OFFSET
#define GPIO_MASK_DATA_5_MSW_OFFSET             0XFF0A002C
#undef GPIO_DIRM_5_OFFSET
#define GPIO_DIRM_5_OFFSET                      0XFF0A0344
#undef GPIO_OEN_5_OFFSET
#define GPIO_OEN_5_OFFSET                       0XFF0A0348
#undef GPIO_DATA_5_OFFSET
#define GPIO_DATA_5_OFFSET                      0XFF0A0054
#define GPIO_DATA_RO_5_OFFSET                   0XFF0A0074

extern int init_si5341(void);

#define UART_BASEADDR XPAR_XUARTPS_0_BASEADDR
#define UART_DEVICE_ID XPAR_XUARTPS_0_DEVICE_ID
#define UART_BAUDRATE 115200

#define DMA0_BUF_ADDR 0x10000000
#define DMA1_BUF_ADDR 0x20000000

#define SYS_FSI_ADDR 0xA0000000

static int pwr_cycle(void);

static XGpioPs Gpio;	/* The driver instance for GPIO Device. */
#define MIO_LED2  30
#define EMIO_RSTN 173

static void
uart_init(void)
{
  u32 reg;

  //  Baudrate
  //  XUartPs_SetBaudRate(UART_BASEADDR, UART_BAUDRATE);

  // Enable UART tx and rx.
  reg = XUartPs_ReadReg(UART_BASEADDR, XUARTPS_CR_OFFSET);
  XUartPs_WriteReg(UART_BASEADDR, XUARTPS_CR_OFFSET,
                   ((reg & ~XUARTPS_CR_EN_DIS_MASK) |
                    XUARTPS_CR_TX_EN | XUARTPS_CR_RX_EN));
}

static void
uart_putc(unsigned c)
{
  if (c == '\n')
    XUartPs_SendByte(UART_BASEADDR, '\r');
  XUartPs_SendByte(UART_BASEADDR, c);
}

static unsigned
uart_getc(void)
{
  return XUartPs_RecvByte(UART_BASEADDR);
}

/* Like uart_getc but for a multi byte key.  Wait the next character for
   at most 1 character transmit time.  */
static unsigned
uart_getc_next(void)
{
  unsigned i;

  /* Wait for (8 bits, 1 start, 1 stop) * 2.  */
  for (i = 0; i < 2*(8 + 2); i++) {
    if (XUartPs_IsReceiveData(UART_BASEADDR)) {
      return XUartPs_RecvByte(UART_BASEADDR);
    }
    usleep ((1000000 + UART_BAUDRATE) / UART_BAUDRATE);
  }
  return 0;
}

static void
uart_puts(const char *s)
{
  while (*s)
    uart_putc(*s++);
}

static void
uart_putpad (unsigned len, unsigned pad, unsigned zero)
{
  while (len < pad)
    {
      uart_putc (zero ? '0' : ' ');
      len++;
    }
}

static void
uart_putnum_uns (unsigned long val, unsigned pad, unsigned zero)
{
  char buf[20];
  char *p;

  buf[19] = 0;
  p = &buf[19];

  do
    {
      unsigned nval = val / 10;
      *--p = '0' + (val - nval * 10);
      val = nval;
    }
  while (val != 0);
  uart_putpad (&buf[19] - p, pad, zero);
  uart_puts (p);
}

static void
uart_putnum_hex (unsigned long val, unsigned pad, unsigned zero)
{
  char buf[17];
  char *p;

  buf[16] = 0;
  p = &buf[16];

  do
    {
      unsigned d = val & 15;
      if (d < 10)
        d += '0';
      else
        d += 'a' - 10;
      *--p = d;
      val >>= 4;
    }
  while (val != 0);
  uart_putpad (&buf[16] - p, pad, zero);
  uart_puts (p);
}

static void uart_printf (const char *fmt, ...) __attribute__((format(printf, 1, 2)));

static void
uart_printf (const char *fmt, ...)
{
  va_list args;
  va_start (args, fmt);

  for (; *fmt; fmt++)
    if (*fmt != '%')
      uart_putc (*fmt);
    else
      {
        unsigned zero = 0;
        unsigned pad = 0;
        unsigned lg = 0;
        unsigned long val;

        fmt++;
        if (*fmt == '0')
          {
            zero = 1;
            fmt++;
          }
        for (; *fmt >= '0' && * fmt <= '9'; fmt++)
          pad = pad * 10 + *fmt - '0';

        for (; *fmt == 'l'; fmt++)
          lg++;

        switch (*fmt)
          {
          case '%':
            uart_putc ('%');
            break;
          case 's':
            uart_puts (va_arg(args, char *));
            break;
          case 'u':
            if (lg)
              val = va_arg(args, unsigned long);
            else
              val = va_arg(args, unsigned);
            uart_putnum_uns (val, pad, zero);
            break;
          case 'x':
            if (lg)
              val = va_arg(args, unsigned long);
            else
              val = va_arg(args, unsigned);
            uart_putnum_hex (val, pad, zero);
            break;
          default:
            uart_putc ('?');
            break;
          }
      }

  va_end (args);
}

#define UART_HIST_MAX 16
#define UART_HIST_LINELEN 32

/* The buffer for the history.  Fully static.  */
static char uart_history[UART_HIST_MAX][UART_HIST_LINELEN];
/* Number of used entries in the history.  Must be <= UART_HIST_MAX.  */
static unsigned uart_hist_len;
/* The next index to add a new entry.  */
static unsigned uart_hist_next;

static void
uart_history_add (const char *line)
{
  unsigned len = strlen(line);
  char *dest = uart_history[uart_hist_next];

  if (len >= UART_HIST_LINELEN) {
    /* Truncate.  */
    len = UART_HIST_LINELEN - 1;
  }
  memcpy (dest, line, len);
  dest[len] = 0;

  uart_hist_next = (uart_hist_next + 1) % UART_HIST_MAX;
  if (uart_hist_len < UART_HIST_MAX)
    uart_hist_len++;
}

struct uart_rl_context {
  char *buf;
  unsigned maxlen;
  unsigned pos;
  unsigned hist_pos;
};

static void
uart_rl_backspace(struct uart_rl_context *ctxt)
{
  if (ctxt->pos == 0)
    uart_putc('\a');
  else
    {
      uart_puts("\b \b");
      ctxt->pos--;
    }
}

static void
uart_rl_clear(struct uart_rl_context *ctxt)
{
  for (; ctxt->pos > 0; ctxt->pos--)
    uart_puts("\b \b");
}

/* Internal function: replace current buffer with the new history position.  */
static void
uart_rl_hist_replace(struct uart_rl_context *ctxt)
{
    /* Select entry.  */
    unsigned hidx =
      (uart_hist_next - ctxt->hist_pos + UART_HIST_MAX) % UART_HIST_MAX;
    const char *h = uart_history[hidx];
    unsigned hlen = strlen(h);
    unsigned i;

    if (hlen > ctxt->maxlen)
      hlen = ctxt->maxlen;

    /* Clear current line.  */
    for (i = 0; i < ctxt->pos; i++)
      uart_putc('\b');

    /* Write line.  */
    for (i = 0; i < hlen; i++) {
      uart_putc(h[i]);
      ctxt->buf[i] = h[i];
    }

    /* Clear extra char.  */
    for (i = hlen; i < ctxt->pos; i++)
      uart_putc(' ');
    for (i = hlen; i < ctxt->pos; i++)
      uart_putc('\b');
    ctxt->pos = hlen;
}

static void
uart_rl_prev_hist(struct uart_rl_context *ctxt)
{
  if (ctxt->hist_pos >= uart_hist_len) {
    /* No more entries.  */
    uart_putc('\a');
    return;
  }

  ctxt->hist_pos++;

  uart_rl_hist_replace(ctxt);
}

static void
uart_rl_next_hist(struct uart_rl_context *ctxt)
{
  if (ctxt->hist_pos == 0) {
    /* No more entries.  */
    uart_putc('\a');
    return;
  }

  if (--ctxt->hist_pos == 0)
    uart_rl_clear(ctxt);
  else
    uart_rl_hist_replace(ctxt);
}

static void
uart_rl_add(struct uart_rl_context *ctxt, char c)
{
  if (c >= ' ' && ctxt->pos < ctxt->maxlen - 1)
    {
      ctxt->buf[ctxt->pos++] = c;
      uart_putc (c);
    }
  else
    uart_putc ('\a');
}

static int
uart_readline(const char *prompt, char *buf, unsigned maxlen)
{
  struct uart_rl_context ctxt;

  ctxt.buf = buf;
  ctxt.maxlen = maxlen;
  ctxt.pos = 0;
  ctxt.hist_pos = 0;

  uart_puts (prompt);

  while (1)
    {
      unsigned c = uart_getc();

      switch(c)
        {
        case '\r':
          uart_putc ('\n');
          /* Remove trailing spaces.  */
          while (ctxt.pos > 0 && ctxt.buf[ctxt.pos - 1] == ' ')
            ctxt.pos--;
          /* Append a NULL.  */
          ctxt.buf[ctxt.pos] = 0;
          return ctxt.pos;

        case '\b':
        case 0x7f:
          /* Backspace or DEL.  */
          uart_rl_backspace(&ctxt);
          break;

        case 'U' - '@':
          /* C-u: clear the whole line.  */
          uart_rl_clear(&ctxt);
          break;

        case 'P' - '@':
          /* C-p: previous command.  */
          uart_rl_prev_hist(&ctxt);
          break;

        case 'N' - '@':
          /* C-n: next command.  */
          uart_rl_next_hist(&ctxt);
          break;

        case '\e':
          /* ESC.  Sequence expected.  */
          {
            unsigned c1 = uart_getc_next();

            if (c1 == '[') {
              unsigned c2 = uart_getc_next();

              if (c2 == 'A')
                uart_rl_prev_hist(&ctxt);
              else if (c2 == 'B')
                uart_rl_next_hist(&ctxt);
              else
                uart_putc('\a');
            }
            else
              uart_putc('\a');
          }
          break;

        default:
          uart_rl_add(&ctxt, c);
          break;
        }
    }
}

static char *
arg_skip(char *l)
{
  while (*l == ' ')
    l++;
  return l;
}

/* Extract next work from LINE.  */
static char *
arg_next(char **line)
{
  char *l = *line;
  char *res;

  l = arg_skip (l);
  if (*l == 0)
    {
      /* End of line.  */
      *line = NULL;
      return NULL;
    }
  res = l;

  while (*l != ' ') {
    if (*l == 0) {
      *line = NULL;
      return res;
    }
    l++;
  }

  *line = l + 1;
  *l = 0;
  return res;
}

static int
strtou (char *l, unsigned *val)
{
  unsigned res;
  char c;

  res = 0;
  if (l[0] == '0' && l[1] == 'x')
    {
      for (l += 2; (c = *l) != 0; l++)
        {
          if (c >= '0' && c <= '9')
            c -= '0';
          else if (c >= 'a' && c <= 'f')
            c -= 'a' - 10;
          else if (c >= 'A' && c <= 'F')
            c -= 'A' - 10;
          else
            return -1;
          res = (res << 4) + c;
        }
    }
  else
    {
      for (; (c = *l) != 0; l++)
        {
          if (c >= '0' && c <= '9')
            c -= '0';
          else
            return -1;
          res = res * 10 + c;
        }
    }
  *val = res;
  return 0;
}

static int
arg_next_number (char **line, unsigned *val)
{
  char *l = arg_next(line);

  if (l == NULL)
    return -1;

  return strtou (l, val);
}

static int
emio_init(void)
{
	XGpioPs_Config *ConfigPtr;
	int status;

	ConfigPtr = XGpioPs_LookupConfig(XPAR_XGPIOPS_0_DEVICE_ID);
	status = XGpioPs_CfgInitialize(&Gpio, ConfigPtr,
			ConfigPtr->BaseAddr);
	if (status != XST_SUCCESS) {
		print("Reset_PL: Init failed\n\r");
		return XST_FAILURE;
	}

        uart_printf("gpio maxpin: %u, maxbank: %u\n",
                    Gpio.MaxPinNum, Gpio.MaxBanks);

	XGpioPs_SetDirectionPin(&Gpio, EMIO_RSTN, 1);
	XGpioPs_SetDirectionPin(&Gpio, MIO_LED2, 1);
	XGpioPs_SetOutputEnablePin(&Gpio, EMIO_RSTN, 1);
	XGpioPs_SetOutputEnablePin(&Gpio, MIO_LED2, 1);

        return 0;
}

static void
do_gpio(char *args)
{
  uart_printf("bank5 mask: %08x\n", Xil_In32(GPIO_MASK_DATA_5_MSW_OFFSET));
  uart_printf("bank5 dirm: %08x\n", Xil_In32(GPIO_DIRM_5_OFFSET));
  uart_printf("bank5 oen:  %08x\n", Xil_In32(GPIO_OEN_5_OFFSET));
  uart_printf("bank5 data: %08x\n", Xil_In32(GPIO_DATA_5_OFFSET));
  uart_printf("bank5 ro:   %08x\n", Xil_In32(GPIO_DATA_RO_5_OFFSET));
}

static unsigned arg_addr = DMA0_BUF_ADDR;
static unsigned arg_len = 256;

static int
arg_memory(const char *name, char *args)
{
  if (args != NULL && arg_next_number(&args, &arg_addr) != 0)
    {
      uart_puts ("address expected\n");
      return -1;
    }
  if (args != NULL && arg_next_number(&args, &arg_len) != 0)
    {
      uart_puts ("length expected\n");
      return -1;
    }
  uart_printf ("%s at 0x%x [%u]:\n", name, arg_addr, arg_len);
  return 0;
}

static void
do_dump(char *args)
{
  unsigned i, j;

  if (arg_memory("dump", args) != 0)
    return;

  for (i = 0; i < arg_len; i += 16)
    {
      uart_printf("%08x:", arg_addr + i);
      for (j = 0; j < 16; j += 4)
        {
          unsigned val = Xil_In32(arg_addr + i + j);
          uart_printf(" %08x", val);
        }
      uart_putc('\n');
    }

  arg_addr += arg_len;
}

static void
dma_init(unsigned nsamp)
{
  Xil_Out32(SYS_FSI_ADDR | SYS_TOP_SUB_REG_CONFIGS
            | SYS_TOP_SUB_REG_CONFIGS_ADDRESS, DMA0_BUF_ADDR);
  Xil_Out32(SYS_FSI_ADDR | SYS_TOP_SUB_REG_CONFIGS
            | SYS_TOP_SUB_REG_CONFIGS_ADDRESS
            | SYS_TOP_SUB_REG_CONFIGS_SIZE, DMA1_BUF_ADDR);

  Xil_Out32(SYS_FSI_ADDR | SYS_TOP_SUB_REG_LENGTH, nsamp);
}

static unsigned
get_dma_length(void)
{
  return Xil_In32(SYS_FSI_ADDR | SYS_TOP_SUB_REG_LENGTH);
}

static void
dma_cache_invalid_with_nsamp(unsigned nsamp)
{
  Xil_DCacheInvalidateRange(DMA0_BUF_ADDR, nsamp * 16 * 3);
  Xil_DCacheInvalidateRange(DMA1_BUF_ADDR, nsamp * 16 * 3);
}

static void
do_dma(char *args)
{
  unsigned status;
  unsigned nsamp = 200;
  unsigned val;

  if (args != NULL) {
    if (arg_next_number(&args, &nsamp) != 0) {
      uart_puts ("nbr samples expected\n");
      return;
    }
    dma_init(nsamp);
  } else {
    nsamp = get_dma_length();
  }

  dma_cache_invalid_with_nsamp(nsamp);

  uart_printf("Starting (%u samps)...\n", nsamp);
  val = Xil_In32(SYS_FSI_ADDR |SYS_TOP_SUB_REG_CONTROL);
  Xil_Out32(SYS_FSI_ADDR |SYS_TOP_SUB_REG_CONTROL,
            val | SYS_TOP_SUB_REG_CONTROL_SW_TRIG);
  do
    {
      usleep (10);
      status = Xil_In32(SYS_FSI_ADDR | SYS_TOP_SUB_REG_STATUS);
    }
  while ((status & SYS_TOP_SUB_REG_STATUS_TRANSFER) != 0);
  uart_printf("Done 0x%08x - 0x%08x\n",
              DMA0_BUF_ADDR, DMA0_BUF_ADDR + nsamp * 16 * 3 - 1);
  dma_cache_invalid_with_nsamp(nsamp);
}

static void
do_regs(char *args)
{
  unsigned i;
  uart_printf("ident:   %08x\n",
              Xil_In32(SYS_FSI_ADDR | SYS_TOP_SUB_REG_IDENT));
  uart_printf("status:  %08x\n",
              Xil_In32(SYS_FSI_ADDR | SYS_TOP_SUB_REG_STATUS));
  uart_printf("control: %08x\n",
              Xil_In32(SYS_FSI_ADDR | SYS_TOP_SUB_REG_CONTROL));
  uart_printf("config:  %08x\n",
              Xil_In32(SYS_FSI_ADDR | SYS_TOP_SUB_REG_CONFIG));
  uart_printf("length:  %08x\n",
              Xil_In32(SYS_FSI_ADDR | SYS_TOP_SUB_REG_LENGTH));
  for (i = 0; i < 2; i++)
    {
      unsigned addr = SYS_FSI_ADDR + SYS_TOP_SUB_REG_CONFIGS
        + (i * SYS_TOP_SUB_REG_CONFIGS_SIZE);
      uart_printf("address  %u: %08x\n", i,
                 Xil_In32(addr + SYS_TOP_SUB_REG_CONFIGS_ADDRESS));
      uart_printf("cur addr %u: %08x\n", i,
                 Xil_In32(addr + SYS_TOP_SUB_REG_CONFIGS_CUR_ADDRESS));
      uart_printf("status   %u: %08x\n", i,
                 Xil_In32(addr + SYS_TOP_SUB_REG_CONFIGS_STATUS));
    }
  uart_printf("stat1:  %08x\n",
              Xil_In32(SYS_FSI_ADDR | SYS_TOP_SUB_REG_PERIPH_STAT1));
  uart_printf("stat2:  %08x\n",
              Xil_In32(SYS_FSI_ADDR | SYS_TOP_SUB_REG_PERIPH_STAT2));
  uart_printf("cmd boards:  %08x\n",
              Xil_In32(SYS_FSI_ADDR | SYS_TOP_SUB_REG_CMD_BOARDS));
  uart_printf("cmd value:   %08x\n",
              Xil_In32(SYS_FSI_ADDR | SYS_TOP_SUB_REG_CMD_VALUE));
  uart_printf("cmd status:  %08x\n",
              Xil_In32(SYS_FSI_ADDR | SYS_TOP_SUB_REG_CMD_STATUS));
  for (i = 0; i < 8; i++)
    uart_printf("cmd reply[%u]: %08x\n", i,
                Xil_In32(SYS_FSI_ADDR | SYS_TOP_SUB_REG_CMD_REPLY | (i * 4)));
}

static void
do_clear(char *args)
{
  unsigned i;

  if (arg_memory("clear", args) != 0)
    return;

  for (i = 0; i < arg_len; i += 4)
    Xil_Out32(arg_addr + i, 0);

  arg_addr += arg_len;
}

static void
do_dcache(char *args)
{
  char *arg;

  arg = arg_next(&args);
  if (arg != NULL && strcmp (arg, "off") == 0) {
    Xil_DCacheFlush();
    Xil_DCacheDisable();
  }
  else if (arg != NULL && strcmp (arg, "on") == 0) {
    Xil_DCacheEnable();
  }
  else {
    uart_puts("missing on/off\n");
    return;
  }
}

/* Read sample number OFF from channel CHANNEL of board BOARD.
   BOARD is 1 based, CHANNEL and OFF are 0 based.  */
static unsigned
read_sample (unsigned board, unsigned channel, unsigned off)
{
  unsigned *samp;
  unsigned b;
  unsigned ch;

  /* Check bounds.  */
  if (board < 1 || board > 8)
    return 0;
  if (channel > 7)
    return 0;

  if (board <= 4)
    samp = (unsigned *)DMA0_BUF_ADDR;
  else
    samp = (unsigned *)DMA1_BUF_ADDR;

  /* Each sample is 1.5 bytes, so 48 (= 4*8*1.5) bytes per bank for all
     channels (8 chan per board, 4 boards per bank).
     48 bytes is 12 words.  */
  samp += 12 * off;

  b = 3 - ((board - 1) & 3);

  /* Extract the data for 1 board (8 ch * 12 bits = 96 bits).  */
  unsigned long hi =
    samp[2 * b] | (((unsigned long)samp[2 * b + 1]) << 32);
  unsigned lo = samp[8 + b];

  /* Extract the sample for the channel.  */
  ch = channel;
  unsigned v = (((hi >> (8 * ch)) & 0xff) << 4) | ((lo >> (4 * ch)) & 0x0f);

  return v;
}

static void
do_samp(char *args)
{
  char *arg;
  unsigned bank = 0;
  unsigned nbr = 32;
  int i, j, k;

  arg = NULL;
  while (args != NULL) {
    arg = arg_next(&args);
    if (arg == NULL)
      break;
    if (arg[0] != '-')
      break;
    if (strcmp(arg, "-0") == 0)
      bank = 0;
    else if (strcmp(arg, "-1") == 0)
      bank = 1;
    else if (strcmp(arg, "-h") == 0) {
      uart_puts ("usage: samp [-0|-1] [NBR]\n");
      return;
    }
    else {
      uart_puts ("bad option, try -h\n");
      return;
    }
    arg = NULL;
  }

  if (arg != NULL && strtou(arg, &nbr) != 0) {
    uart_puts ("nbr samples expected\n");
    return;
  }
  else
    arg = arg_next(&args);

  if (arg != NULL) {
    uart_puts ("extra argument; try -h\n");
    return;
  }

  for (i = 0; i < nbr; i++)
    {
      for (j = 0; j < 4; j++)
        {
          if ((j % 2) == 0)
            uart_printf ("%03u+%u:", i, j);

          for (k = 0; k < 8; k++)
            uart_printf (" %03x", read_sample (4 * bank + j + 1, k, i));
          if ((j % 2) == 1)
            uart_putc ('\n');
          else
            uart_puts (" -");
        }
    }
}

static void
do_graph(char *args)
{
  char *arg;
  unsigned *samp;
  unsigned i, j, k;
  unsigned char line[65];
  unsigned nsamp = 50;
  unsigned step = 1;

  samp = (unsigned *)DMA0_BUF_ADDR;

  arg = NULL;
  while (args != NULL) {
    arg = arg_next(&args);
    if (arg == NULL)
      break;
    if (arg[0] != '-')
      break;
    if (strcmp(arg, "-0") == 0)
      samp = (unsigned *)DMA0_BUF_ADDR;
    else if (strcmp(arg, "-1") == 0)
      samp = (unsigned *)DMA1_BUF_ADDR;
    else if (strcmp(arg, "-h") == 0) {
      uart_puts ("usage: graph [-0|-1] [NSAMP [STEP]]\n");
      return;
    }
    else {
      uart_puts ("bad option, try -h\n");
      return;
    }
    arg = NULL;
  }

  if (arg != NULL && strtou(arg, &nsamp) != 0)
    {
      uart_puts ("nbr samples expected\n");
      return;
    }
  else
    arg = arg_next(&args);
  if (arg != NULL && strtou(arg, &step) != 0)
    {
      uart_puts ("step expected\n");
      return;
    }
  else
    arg = arg_next(&args);
  if (arg != NULL) {
    uart_puts ("extra argument; try -h\n");
    return;
  }

  dma_cache_invalid_with_nsamp(nsamp);

  for (i = 0; i < nsamp; i += step)
    {
      for (j = 0; j < 64; j++)
        line[j] = ' ';
      line[64] = 0;

      for (j = 0; j < 4; j++)
        {
          unsigned long hi =
            samp[2*j] | (((unsigned long)samp[2*j + 1]) << 32);

          for (k = 0; k < 8; k++)
            {
              unsigned v = (hi >> (2 + 8 * k)) & 0x3f;
              unsigned n = k + j * 8;

              if (line[v] == ' ')
                line[v] = n < 26 ? 'a' + n : 'A' + n - 26;
              else
                line[v] = '*';
            }
        }

      uart_printf ("%04u: %s\n", i, line);

      samp += 12 * step;
    }
}

static int
parse_ch_mask(const char *s, unsigned *res)
{
  *res = 0;

  /* Very simple parsing.  */
  while (*s) {
    if (*s >= '0' && *s <= '7')
      *res |= 1 << (*s - '0');
    else if (*s == ',')
      ;
    else {
      uart_puts("bad character in channels mask\n");
      return -1;
    }
    s++;
  }
  return 0;
}

static void
do_raw(char *args)
{
  unsigned board = 1;
  unsigned nbr = 32;
  unsigned mask = 1;
  int i;

  if (args != NULL && arg_next_number(&args, &board) != 0)
    {
      uart_puts ("board expected\n");
      return;
    }
  if (args != NULL)
    {
      char *arg = arg_next(&args);
      if (arg == NULL) {
        uart_puts ("channel expected\n");
        return;
      }
      if (parse_ch_mask(arg, &mask) != 0)
        return;
    }
  if (args != NULL && arg_next_number(&args, &nbr) != 0)
    {
      uart_puts ("length expected\n");
      return;
    }

  dma_cache_invalid_with_nsamp(nbr);

  for (i = 0; i < nbr; i++) {
    unsigned m = mask;

    for (unsigned j = 0; j < 8; j++) {
      if (m & (1 << j)) {
        uart_printf ("%03x", read_sample (board, j, i));
        m &= ~(1 << j);
        if (m == 0)
          break;
        uart_putc(',');
      }
    }
    uart_putc('\n');
  }
}

static void
do_config(char *args)
{
  unsigned val;
  char *arg;

  val = Xil_In32(SYS_FSI_ADDR | SYS_TOP_SUB_REG_CONFIG);

  while (1) {
    arg = arg_next(&args);
    if (arg == NULL)
      break;
    if (strcmp (arg, "+patterns") == 0)
      val |= SYS_TOP_SUB_REG_CONFIG_PATTERNS;
    else if (strcmp (arg, "-patterns") == 0)
      val &= ~SYS_TOP_SUB_REG_CONFIG_PATTERNS;
    else if (strcmp (arg, "resync") == 0)
      val |= SYS_TOP_SUB_REG_CONFIG_RESYNC;
    else if (strcmp (arg, "+trig") == 0)
      val |= SYS_TOP_SUB_REG_CONFIG_TRIG_EN;
    else if (strcmp (arg, "-trig") == 0)
      val &= ~SYS_TOP_SUB_REG_CONFIG_TRIG_EN;
    else if (strcmp (arg, "+rst") == 0)
      val |= SYS_TOP_SUB_REG_CONFIG_RST_N;
    else if (strcmp (arg, "-rst") == 0)
      val &= ~SYS_TOP_SUB_REG_CONFIG_RST_N;
    else if (strcmp (arg, "+clk") == 0)
      val |= SYS_TOP_SUB_REG_CONFIG_CLK_SEL_MASK;
    else if (strcmp (arg, "-clk") == 0)
      val &= ~SYS_TOP_SUB_REG_CONFIG_CLK_SEL_MASK;
    else {
      uart_printf("unknown command %s, try +/-patterns, resync, +/-trig\n",
                  arg);
      return;
    }
  }
  uart_printf ("config: %08x, boards-off: %02x, trig-en: %x\n",
               val,
               (unsigned)((val & SYS_TOP_SUB_REG_CONFIG_PERIPH_OFF_MASK)
                >> SYS_TOP_SUB_REG_CONFIG_PERIPH_OFF_SHIFT),
               !!(val & SYS_TOP_SUB_REG_CONFIG_TRIG_EN));
  Xil_Out32(SYS_FSI_ADDR | SYS_TOP_SUB_REG_CONFIG, val);
}

static void
fsi_reset(void)
{
  unsigned val;

  val = Xil_In32(SYS_FSI_ADDR | SYS_TOP_SUB_REG_CONFIG);
  /* Enable clock, put 0 on rst_n  */
  val |= SYS_TOP_SUB_REG_CONFIG_CLK_SEL_MASK;
  val &= ~SYS_TOP_SUB_REG_CONFIG_RST_N;
  Xil_Out32(SYS_FSI_ADDR | SYS_TOP_SUB_REG_CONFIG, val);
  Xil_Out32(SYS_FSI_ADDR | SYS_TOP_SUB_REG_CONFIG, val);
  /* Release rst_n  */
  val |= SYS_TOP_SUB_REG_CONFIG_RST_N;
  Xil_Out32(SYS_FSI_ADDR | SYS_TOP_SUB_REG_CONFIG, val);
}

static void
fsi_trig_enable(unsigned en)
{
  unsigned val;

  val = Xil_In32(SYS_FSI_ADDR | SYS_TOP_SUB_REG_CONFIG);
  if (en)
    val |= SYS_TOP_SUB_REG_CONFIG_TRIG_EN;
  else
    val &= ~SYS_TOP_SUB_REG_CONFIG_TRIG_EN;
  Xil_Out32(SYS_FSI_ADDR | SYS_TOP_SUB_REG_CONFIG, val);
}

static void
fsi_wait_cmd(void)
{
  unsigned mask;
  unsigned i;

  mask = Xil_In32(SYS_FSI_ADDR | SYS_TOP_SUB_REG_CMD_BOARDS);

  for (i = 0; i < 4; i++) {
    unsigned status = Xil_In32(SYS_FSI_ADDR | SYS_TOP_SUB_REG_CMD_STATUS);
    unsigned done = (status >> SYS_TOP_SUB_REG_CMD_STATUS_DONE_SHIFT);
    if ((done & mask) == mask)
      return;
  }

  /* Could happen if a board is sending samples.  */
  uart_puts("cmd timeout\n");
}

static void
fsi_wait_spi(unsigned mask)
{
  unsigned i = 0;

  for (i = 0; i < 100; i++) {
    unsigned s1 = Xil_In32(SYS_FSI_ADDR | SYS_TOP_SUB_REG_PERIPH_STAT1);
    unsigned busy = s1 >> SYS_TOP_SUB_REG_PERIPH_STAT1_SPI_BUSY_SHIFT;
    if (!(busy & mask))
      return;
  }

  /* Could happen if a board is sending samples.  */
  uart_puts("spi timeout\n");
}

/* Return the mask of existing boards.  */
static unsigned
fsi_get_boards_mask(void)
{
  unsigned s1 = Xil_In32(SYS_FSI_ADDR | SYS_TOP_SUB_REG_PERIPH_STAT1);
  return~s1 & SYS_TOP_SUB_REG_PERIPH_STAT1_INP_MASK;
}

/* Select boards for the commands.  */
static void
fsi_sel_boards(unsigned boards)
{
  Xil_Out32(SYS_FSI_ADDR | SYS_TOP_SUB_REG_CMD_BOARDS, boards);
}

static void
fsi_send_value(unsigned val)
{
  Xil_Out32(SYS_FSI_ADDR | SYS_TOP_SUB_REG_CMD_VALUE, val);
  Xil_Out32(SYS_FSI_ADDR | SYS_TOP_SUB_REG_CMD_CTRL,
            SYS_TOP_SUB_REG_CMD_CTRL_SEND);
  fsi_wait_cmd();
}

static void
fsi_exec_cmd(unsigned cmd)
{
  Xil_Out32(SYS_FSI_ADDR | SYS_TOP_SUB_REG_CMD_VALUE, cmd);
  Xil_Out32(SYS_FSI_ADDR | SYS_TOP_SUB_REG_CMD_CTRL,
            SYS_TOP_SUB_REG_CMD_CTRL_EXEC);
  fsi_wait_cmd();
}

/* Get last reply from fsi SLOT (0-7).  */
static unsigned
fsi_get_reply(unsigned slot)
{
  return Xil_In32(SYS_FSI_ADDR | SYS_TOP_SUB_REG_CMD_REPLY | (slot * 4));
}

static void
fsi_write_spi(unsigned val)
{
  unsigned boards = fsi_get_boards_mask();
  fsi_sel_boards(boards);

  fsi_send_value(val);

  /* Send command to adc.  */
  unsigned cmd = 0x5c;
  fsi_exec_cmd(cmd);

  /* Wait until end of SPI transfer. */
  fsi_wait_spi(boards);
}

static unsigned long
fsi_read_clock(void)
{
  unsigned hi, hi1, lo;

  hi = Xil_In32(SYS_FSI_ADDR | SYS_TOP_SUB_REG_STAT_CLOCK | 4);
  while (1) {
    lo = Xil_In32(SYS_FSI_ADDR | SYS_TOP_SUB_REG_STAT_CLOCK);
    hi1 = Xil_In32(SYS_FSI_ADDR | SYS_TOP_SUB_REG_STAT_CLOCK | 4);
    if (hi1 == hi)
      break;
    hi = hi1;
  }
  return (((unsigned long)hi) << 32) | (unsigned long)lo;
}

static void
do_fsi_delay(char *args, unsigned delay_sel)
{
  char *arg;
  unsigned val;

  unsigned board;
  if (arg_next_number(&args, &board) != 0
      || !(board >= 1 && board <= 8)) {
    uart_puts("missing or bad board number\n");
    return;
  }

  unsigned lane;
  if (arg_next_number(&args, &lane) != 0
      || !(lane >= 0 && lane <= 12)) {
    uart_puts("missing or bad lane number\n");
    return;
  }

  if (args == NULL) {
    uart_puts("missing inc/dec option\n");
    return;
  }

  unsigned sel = ((board - 1) << SYS_TOP_SUB_REG_STAT_SEL_BOARD_SHIFT)
    | (lane << SYS_TOP_SUB_REG_STAT_SEL_CH_SHIFT)
    | SYS_TOP_SUB_REG_STAT_SEL_BRANCH;
  Xil_Out32(SYS_FSI_ADDR | SYS_TOP_SUB_REG_STAT_SEL, sel);

  arg = arg_next(&args);

  val = Xil_In32(SYS_FSI_ADDR | SYS_TOP_SUB_REG_DELAY_CTRL);
  if (strcmp(arg, "inc") == 0)
    val |= SYS_TOP_SUB_REG_DELAY_CTRL_INC;
  else if (strcmp(arg, "dec") == 0)
    val |= SYS_TOP_SUB_REG_DELAY_CTRL_DEC;
  else {
    uart_puts("bad delay command\n");
    return;
  }
  Xil_Out32(SYS_FSI_ADDR | SYS_TOP_SUB_REG_DELAY_CTRL, val);
}

static void
do_fsi(char *args)
{
  char *arg;

  arg = arg_next(&args);
  if (arg == NULL) {
    uart_puts("missing subcommand\n");
    return;
  }

  if (strcmp (arg, "rst") == 0) {
    fsi_reset();
  }
  else if (strcmp (arg, "status") == 0) {
    unsigned s1 = Xil_In32(SYS_FSI_ADDR | SYS_TOP_SUB_REG_PERIPH_STAT1);
    unsigned s2 = Xil_In32(SYS_FSI_ADDR | SYS_TOP_SUB_REG_PERIPH_STAT2);
    unsigned mask = ~s1 & SYS_TOP_SUB_REG_PERIPH_STAT1_INP_MASK;
    unsigned i;

    uart_printf ("stat1: %08x, stat2: %08x\n", s1, s2);
    uart_printf ("spi-busy: %02x, trigger: %02x, presence: %02x\n",
                 (s1 >> SYS_TOP_SUB_REG_PERIPH_STAT1_SPI_BUSY_SHIFT) & mask,
                 (s1 >> SYS_TOP_SUB_REG_PERIPH_STAT1_TRIGGERS_SHIFT) & mask,
                 mask);

    for (i = 0; i < 8; i++) {
      if (!(mask & (1 << i))) {
        /* Board not present.  */
        continue;
      }
      uart_printf ("slot %u: ", i + 1);

      if (!(s2 & (1 << i)))
        uart_putc ('!');
      uart_puts("rx-locked ");
      if (s2 & (1 << i)) {
        if (!(s2 & (1 << (8 + i))))
          uart_putc ('!');
        uart_puts("adc-locked ");

        if (!(s2 & (1 << (16 + i))))
          uart_putc ('!');
        uart_puts("sys-locked ");

        if (!(s2 & (1 << (24 + i))))
          uart_putc ('!');
        uart_puts("fsi-ready");
      }
      uart_putc('\n');
    }
  }
  else if (strcmp (arg, "normal") == 0) {
    fsi_write_spi (0x25000002);
  }
  else if (strcmp (arg, "ramp") == 0) {
    fsi_write_spi (0x25000042);
  }
  else if (strcmp (arg, "off") == 0) {
    fsi_trig_enable(0);
  }
  else if (strcmp (arg, "on") == 0) {
    fsi_trig_enable(1);
  }
  else if (strcmp (arg, "length") == 0) {
    if (args == NULL) {
      uart_printf("current length: %u\n", get_dma_length());
    }
    else {
      unsigned val;
      if (arg_next_number(&args, &val) != 0 || val == 0) {
        uart_puts("bad length number\n");
      }
      else {
        Xil_Out32(SYS_FSI_ADDR | SYS_TOP_SUB_REG_LENGTH, val);
      }
    }
  }
  else if (strcmp (arg, "stat") == 0) {
    if (args == NULL) {
      unsigned long clk_ms = fsi_read_clock() / 100000;

      uart_printf ("clock: %lu sec + %lu msec\n",
                   clk_ms / 1000, clk_ms % 1000);

      unsigned s1 = Xil_In32(SYS_FSI_ADDR | SYS_TOP_SUB_REG_PERIPH_STAT1);
      unsigned mask = ~s1 & SYS_TOP_SUB_REG_PERIPH_STAT1_INP_MASK;

      for (unsigned i = 0; i < 8; i++) {
        if (!(mask & (1 << i))) {
          /* Board not present.  */
          continue;
        }
        for (unsigned j = 0; j < 13; j++) {
          unsigned v;
          unsigned sel;

          sel = (i << SYS_TOP_SUB_REG_STAT_SEL_BOARD_SHIFT)
            | (j << SYS_TOP_SUB_REG_STAT_SEL_CH_SHIFT);
          Xil_Out32(SYS_FSI_ADDR | SYS_TOP_SUB_REG_STAT_SEL, sel);
          v = Xil_In32(SYS_FSI_ADDR | SYS_TOP_SUB_REG_STAT_BYTE_LOCK);
          uart_printf("slot %u, la %02u: byte:%08x ", i + 1, j, v);
          v = Xil_In32(SYS_FSI_ADDR | SYS_TOP_SUB_REG_STAT_BIT_LOCK);
          uart_printf("bit:%08x ", v);
          v = Xil_In32(SYS_FSI_ADDR | SYS_TOP_SUB_REG_STAT_PRBS);
          uart_printf("prbs:%08x -", v);
          for (unsigned k = 0; k < 4; k++) {
            Xil_Out32(SYS_FSI_ADDR | SYS_TOP_SUB_REG_STAT_SEL,
                      sel | (k << 8));
            v = Xil_In32(SYS_FSI_ADDR | SYS_TOP_SUB_REG_DELAY_CNTVALUE);
            uart_printf (" %03u", v);
          }
          uart_putc('\n');
        }
      }
      return;
    }

    arg = arg_next(&args);
    if (strcmp(arg, "clear") == 0) {
      Xil_Out32(SYS_FSI_ADDR | SYS_TOP_SUB_REG_STAT_SEL,
                SYS_TOP_SUB_REG_STAT_SEL_CLEAR);
    }
    else {
      uart_puts("bad stat command\n");
    }
  }
  else if (strcmp (arg, "resync") == 0) {
    unsigned val;
    val = Xil_In32(SYS_FSI_ADDR | SYS_TOP_SUB_REG_CONFIG);
    val |= SYS_TOP_SUB_REG_CONFIG_RESYNC;
    Xil_Out32(SYS_FSI_ADDR | SYS_TOP_SUB_REG_CONFIG, val);
  }
  else if (strcmp (arg, "vtc") == 0) {
    unsigned val;

    val = Xil_In32(SYS_FSI_ADDR | SYS_TOP_SUB_REG_DELAY_CTRL);
    if (args == NULL) {
      uart_printf ("delay ctrl: %04x ", val);
      if (!(val & SYS_TOP_SUB_REG_DELAY_CTRL_VTC_EN))
        uart_putc('!');
      uart_puts("vtc");
      uart_putc('\n');
      return;
    }

    arg = arg_next(&args);
    if (strcmp(arg, "on") == 0)
      val |= SYS_TOP_SUB_REG_DELAY_CTRL_VTC_EN;
    else if (strcmp(arg, "off") == 0)
      val &= ~SYS_TOP_SUB_REG_DELAY_CTRL_VTC_EN;
    else {
      uart_puts("bad vtc command\n");
      return;
    }
    Xil_Out32(SYS_FSI_ADDR | SYS_TOP_SUB_REG_DELAY_CTRL, val);
  }
  else if (strcmp (arg, "freeze") == 0) {
    unsigned val;

    val = Xil_In32(SYS_FSI_ADDR | SYS_TOP_SUB_REG_DELAY_CTRL);
    if (args == NULL) {
      uart_printf ("delay ctrl: %04x ", val);
      if (!(val & SYS_TOP_SUB_REG_DELAY_CTRL_FREEZE))
        uart_putc('!');
      uart_puts("freeze");
      uart_putc('\n');
      return;
    }

    arg = arg_next(&args);
    if (strcmp(arg, "on") == 0)
      val |= SYS_TOP_SUB_REG_DELAY_CTRL_FREEZE;
    else if (strcmp(arg, "off") == 0)
      val &= ~SYS_TOP_SUB_REG_DELAY_CTRL_FREEZE;
    else {
      uart_puts("bad freeze command\n");
      return;
    }
    Xil_Out32(SYS_FSI_ADDR | SYS_TOP_SUB_REG_DELAY_CTRL, val);
  }
  else if (strcmp (arg, "sdelay") == 0) {
    do_fsi_delay(args, SYS_TOP_SUB_REG_STAT_SEL_BRANCH);
    return;
  }
  else if (strcmp (arg, "ddelay") == 0) {
    do_fsi_delay(args, 0);
    return;
  }
  else if (strcmp (arg, "prbs") == 0) {
    unsigned val;

    val = Xil_In32(SYS_FSI_ADDR | SYS_TOP_SUB_REG_CONTROL);
    if (args == NULL) {
      uart_printf ("control: %04x ", val);
      if (!(val & SYS_TOP_SUB_REG_CONTROL_PRBS))
        uart_putc('!');
      uart_puts("prbs");
      uart_putc('\n');
      return;
    }

    arg = arg_next(&args);
    if (strcmp(arg, "on") == 0)
      val |= SYS_TOP_SUB_REG_CONTROL_PRBS;
    else if (strcmp(arg, "off") == 0)
      val &= ~SYS_TOP_SUB_REG_CONTROL_PRBS;
    else {
      uart_puts("bad prbs command\n");
      return;
    }
    Xil_Out32(SYS_FSI_ADDR | SYS_TOP_SUB_REG_CONTROL, val);
  }
  else if (strcmp (arg, "afe") == 0) {
    if (args == NULL) {
      unsigned boards = fsi_get_boards_mask();

      fsi_exec_cmd(0x59);
      fsi_exec_cmd(0x54);

      for (unsigned i = 0; i < 8; i++) {
        if (!(boards & (1 << i))) {
          /* Board not present.  */
          continue;
        }
        uart_printf("slot %u: afe=%02x\n", i + 1, fsi_get_reply(i));
      }
    }
    else {
      unsigned brd;
      unsigned val;

      if (arg_next_number(&args, &brd) < 0
          || (brd < 1 || brd > 8)) {
        uart_puts ("incorrect board number\n");
        return;
      }
      if (arg_next_number(&args, &val) < 0) {
        uart_puts ("incorrect value\n");
        return;
      }
      fsi_sel_boards (1 << (brd - 1));
      fsi_send_value (val);
      fsi_exec_cmd (0x56);
    }
    return;
  }
  else {
    uart_printf("unknown sub command %s\n", arg);
    uart_puts("usage:\n");
    uart_puts(" rst            reset periph boards\n");
    uart_puts(" status         disp board status\n");
    uart_puts(" ramp           ADC send ramp pattern\n");
    uart_puts(" normal         ADC send data\n");
    uart_puts(" on/off         enable/disable trigger\n");
    uart_puts(" length [NUM]   disp/set number of samples per trigger\n");
    uart_puts(" afe [BOARD VAL]   disp/set analog front-end gain\n");
    uart_puts(" stat [clear]   disp (or clear) statistics\n");
    uart_puts(" resync         resync iserdes bit slip\n");
    uart_puts(" vtc [on|off]   enable/disable iodelay vtc\n");
    uart_puts(" freeze [on|off]   enable/disable iodelay adjustments\n");
    uart_puts(" ddelay BRD LANE inc|dec   adjust data iserdes odelay\n");
    uart_puts(" sdelay BRD LANE inc|dec   adjust sampling iserdes odelay\n");
 }
}

static void
do_led2 (char *args)
{
  unsigned val;

  val = XGpioPs_ReadPin(&Gpio, MIO_LED2);
  XGpioPs_WritePin(&Gpio, MIO_LED2, !val);
  uart_printf("led2: %u\n", !val);
}

static void
do_reset (char *args)
{
  uart_puts("Reset...");
  Xil_Out32(GPIO_MASK_DATA_5_MSW_OFFSET, 0x7fff0000U);
  usleep(500000);
  Xil_Out32(GPIO_MASK_DATA_5_MSW_OFFSET, 0x7fff8000U);
  uart_puts("Done\n");
}

static void
do_rd (char *args)
{
  unsigned addr;
  unsigned val;

  if (arg_next_number(&args, &addr) != 0)
    {
      uart_puts ("address expected\n");
      return;
    }

  val = Xil_In32(addr);
  uart_printf("[%08x] = %08x\n", addr, val);
}

static void
do_wr (char *args)
{
  unsigned addr;
  unsigned val;

  if (arg_next_number(&args, &addr) != 0)
    {
      uart_puts ("address expected\n");
      return;
    }
  if (arg_next_number(&args, &val) != 0)
    {
      uart_puts ("value expected\n");
      return;
    }

  uart_printf("[%08x] <= %08x\n", addr, val);
  Xil_Out32(addr, val);
}

static void
do_cmd(char *args)
{
  char *arg;

  arg = arg_next(&args);
  if (arg == NULL) {
    uart_puts ("usage: boards, value, exec, status, reply\n");
    return;
  }
  if (strcmp (arg, "boards") == 0) {
    unsigned val;
    if (args == NULL)
      val = fsi_get_boards_mask();
    else {
      if (arg_next_number(&args, &val) != 0) {
        uart_puts ("value expected\n");
        return;
      }
    }
    fsi_sel_boards (val);
    return;
  }
  else if (strcmp (arg, "value") == 0) {
    unsigned val;
    if (arg_next_number(&args, &val) != 0) {
      uart_puts ("value expected\n");
      return;
    }
    fsi_send_value(val);
    return;
  }
  else if (strcmp (arg, "exec") == 0) {
    unsigned val;
    if (arg_next_number(&args, &val) != 0) {
      uart_puts ("value expected\n");
      return;
    }
    fsi_exec_cmd(val);
    return;
  }
  else if (strcmp (arg, "status") == 0) {
    unsigned val;
    val = Xil_In32(SYS_FSI_ADDR | SYS_TOP_SUB_REG_CMD_STATUS);
    uart_printf("cmd status: %08x\n", val);
    return;
  }
  else if (strcmp (arg, "reply") == 0) {
    unsigned val;
    int i;
    for (i = 0; i < 8; i++) {
      val = fsi_get_reply(i);
      uart_printf("cmd reply[%u]: %08x\n", i, val);
    }
    return;
  }
  else
    uart_puts ("unknown cmd sub-command\n");
}

static void
do_keys(char *args)
{
  uart_puts("exit by pressing space\n");

  while (1) {
    unsigned c = uart_getc();
    uart_printf ("%02x ", c);
    if (c == ' ')
      break;
  }
  uart_putc('\n');
}

/* Cf https://www.xilinx.com/html_docs/registers/ug1087/mod___uart.html#
   Note: Freq = 100Mhz  */
static void
do_baud(char *args)
{
  unsigned baud;

  if (arg_next_number(&args, &baud) < 0) {
    uart_printf ("base addr: %08x\n", UART_BASEADDR);
    uart_printf ("control:   %08x\n", Xil_In32(UART_BASEADDR + 0x00));
    uart_printf ("mode:      %08x\n", Xil_In32(UART_BASEADDR + 0x04));
    uart_printf ("bd gen:    %08x\n", Xil_In32(UART_BASEADDR + 0x18));
    uart_printf ("bd div:    %08x\n", Xil_In32(UART_BASEADDR + 0x34));
  }
  else {
    unsigned cd;
    unsigned bdiv;
    if (baud == 115200) {
      cd = 124;
      bdiv = 6;
    }
    else if (baud == 230400) {
      cd = 62;
      bdiv = 6;
    }
    else if (baud == 460800) {
      cd = 31;
      bdiv = 6;
    }
    else if (baud == 921600) {
      cd = 18;
      bdiv = 5;
    }
    else {
      uart_puts ("unhandled baudrate\n");
      return;
    }
    Xil_Out32(UART_BASEADDR + 0x18, cd);
    Xil_Out32(UART_BASEADDR + 0x34, bdiv);
  }
}

static void
do_history(char *args)
{
  unsigned i0;

  for (i0 = 0; i0 < uart_hist_len; i0++) {
    unsigned i = uart_hist_len - i0;
    unsigned hidx = (uart_hist_next - i + 2 * UART_HIST_MAX)
      % UART_HIST_MAX;
    uart_printf ("%2u  %s\n", i, uart_history[hidx]);
  }
}

static void do_help(char *args);

struct menu_item
{
  const char *name;
  void (*cmd)(char *args);
};

static const struct menu_item top_menu[] =
  {
   { "led2", do_led2 },
   { "dma", do_dma },
   { "dump", do_dump },
   { "samp", do_samp },
   { "graph", do_graph },
   { "raw", do_raw },
   { "clear", do_clear },
   { "dcache", do_dcache },
   { "regs", do_regs },
   { "config", do_config },
   { "reset", do_reset },
   { "gpio", do_gpio },
   { "fsi", do_fsi },
   { "cmd", do_cmd },
   { "rd", do_rd },
   { "wr", do_wr },
   { "keys", do_keys },
   { "baud", do_baud },
   { "history", do_history },
   { "help", do_help },
   { NULL, NULL }
  };

static void
do_help(char *args)
{
  const struct menu_item *m;

  uart_puts("commands:\n");
  for (m = top_menu; m->name; m++)
    uart_printf (" %s\n", m->name);
}

static void
menu_execute (char *line, const struct menu_item *menu)
{
  char *cmd;

  /* Skip leading blanks.  */
  line = arg_skip(line);

  /* If the line is empty, nothing to do.  */
  if (*line == 0)
    return;

  /* Add to history.  */
  uart_history_add(line);

  cmd = arg_next (&line);

  for (; menu->name; menu++)
    if (strcmp(menu->name, cmd) == 0)
      {
        menu->cmd(line);
        return;
      }
  uart_printf("unknown command '%s'\n", cmd);
}

static int pwr_cycle(void)
{
	XGpioPs Gpio;	/* The driver instance for GPIO Device. */
	int i;

	u32 emio_led = 30;
	u32 emio_pwr = 79;
	u32 emio_pwr_str = 80;

	uart_puts("Power-cycling DI/OT crate\n");

	XGpioPs_SetDirectionPin(&Gpio, emio_led, 1);
	XGpioPs_SetDirectionPin(&Gpio, emio_pwr, 1);
	XGpioPs_SetDirectionPin(&Gpio, emio_pwr_str, 1);
	XGpioPs_SetOutputEnablePin(&Gpio, emio_led, 1);
	XGpioPs_SetOutputEnablePin(&Gpio, emio_pwr, 1);
	XGpioPs_SetOutputEnablePin(&Gpio, emio_pwr, 1);

	XGpioPs_WritePin(&Gpio, emio_pwr, 0x1);
	XGpioPs_WritePin(&Gpio, emio_pwr_str, 0x0);
	XGpioPs_WritePin(&Gpio, emio_pwr_str, 0x1);
	usleep(1000);
	XGpioPs_WritePin(&Gpio, emio_pwr_str, 0x0);
	XGpioPs_WritePin(&Gpio, emio_pwr, 0x0);
	XGpioPs_WritePin(&Gpio, emio_pwr_str, 0x1);
	usleep(1000);
	XGpioPs_WritePin(&Gpio, emio_pwr_str, 0x0);
	usleep(500000);

	uart_puts("Done?\n");

	for (i=0; i<10; ++i) {
		XGpioPs_WritePin(&Gpio, emio_led, 0x1);
		usleep(500000);
		XGpioPs_WritePin(&Gpio, emio_led, 0x0);
		usleep(500000);
	}

	return XST_SUCCESS;
}

int
main(void)
{

  /* Set baudrate to 230400. */
  Xil_Out32(UART_BASEADDR + 0x18, 62);
  Xil_Out32(UART_BASEADDR + 0x34, 6);

	uart_puts("Hello World (" __DATE__ " - " __TIME__ ")\n");
        init_si5341();

	uart_puts("Si5341 config done\n");

        emio_init ();

        uart_init();
        fsi_reset();

        /* 20ms */
        dma_init(2500000);

        fsi_trig_enable(1);

	while (1)
          {
            char buf[32];

            uart_readline("cmd> ", buf, sizeof (buf));
            menu_execute (buf, top_menu);
          }

	return 0;
}
