/*
 * helloworld.c: simple test application
 *
 * This application configures UART 16550 to baud rate 9600.
 * PS7 UART (Zynq) is not initialized by this application, since
 * bootrom/bsp configures it to baud rate 115200
 *
 * ------------------------------------------------
 * | UART TYPE   BAUD RATE                        |
 * ------------------------------------------------
 *   uartns550   9600
 *   uartlite    Configurable only in HW design
 *   ps7_uart    115200 (configured by bootrom/bsp)
 */

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include "xil_printf.h"
#include "xgpiops.h"
#include "xparameters.h"
#include <xil_types.h>
#include <xiicps.h>
#include <xiicps_hw.h>
#include <xuartps_hw.h>

#include "Si5341-RevD-diot-Registers.h"

#define IIC_DEVICE_ID XPAR_PSU_I2C_1_DEVICE_ID
#define IIC_SCLK_RATE       100000
#define I2C_MUX_ADDR        0x70
#define I2C_MUX_CH_SI5341	2
#define I2C_MUX_CH_SFP    5
#define I2C_SI5341_ADDR     0x76

static s32
IicPsSlaveMonitor(XIicPs *InstancePtr, u16 Address)
{
	u32 Index, IntrStatusReg;
	XIicPs *IicPtr;

	IicPtr = InstancePtr;
	XIicPs_EnableSlaveMonitor(InstancePtr, Address);

	Index = 0;
	/*
	 * Wait for the Slave Monitor Status
	 */
	while (Index < 0x00FFFFFF) {
		Index++;
		/*
		 * Read the Interrupt status register.
		 */
		IntrStatusReg = XIicPs_ReadReg(IicPtr->Config.BaseAddress,
				(u32)XIICPS_ISR_OFFSET);
		if (0U != (IntrStatusReg & XIICPS_IXR_SLV_RDY_MASK)) {
			XIicPs_DisableSlaveMonitor(InstancePtr);
			XIicPs_WriteReg(IicPtr->Config.BaseAddress, (u32)XIICPS_ISR_OFFSET,
					IntrStatusReg);
			return XST_SUCCESS;
		}
	}
	XIicPs_DisableSlaveMonitor(InstancePtr);
	return XST_FAILURE;
}

static void
i2cmux_set_channel(XIicPs *InstancePtr, uint8_t channel)
{
	uint8_t buffer[1];

	buffer[0] = 1 << channel;

	XIicPs_MasterSendPolled(InstancePtr, buffer, 1, I2C_MUX_ADDR);
	while (XIicPs_BusIsBusy(InstancePtr))
		;
}

static void
si5341_i2c_write(XIicPs *InstancePtr, uint16_t addr, uint8_t data)
{

	uint8_t buffer[4];

	buffer[0] = 0x01;
	buffer[1] = (0xFF00 & addr) >> 8;
	buffer[2] = (0x00FF & addr);
	buffer[3] = data;

	XIicPs_MasterSendPolled(InstancePtr, buffer, 2, I2C_SI5341_ADDR);
	XIicPs_MasterSendPolled(InstancePtr, &buffer[2], 2, I2C_SI5341_ADDR);
}

static void
si5341_dev_ready(XIicPs *InstancePtr)
{
	uint8_t buffer[4];

	buffer[0] = 0x01;
	buffer[1] = 0;
	buffer[2] = 0xFE;	//DEVICE_READY reg

	XIicPs_MasterSendPolled(InstancePtr, buffer, 2, I2C_SI5341_ADDR);
	XIicPs_MasterSendPolled(InstancePtr, &buffer[2], 1, I2C_SI5341_ADDR);
	do {
		buffer[0] = 0;
		XIicPs_MasterRecvPolled(InstancePtr, buffer, 1, I2C_SI5341_ADDR);
	} while (*buffer != 0xF);
}


static void
si5341_soft_rst(XIicPs *InstancePtr)
{
	uint8_t buffer[4];

	buffer[0] = 0x01;
	buffer[1] = 0;
	buffer[2] = 0x1C;	//SOFT_RST
	buffer[3] = 1;

	XIicPs_MasterSendPolled(InstancePtr, buffer, 2, I2C_SI5341_ADDR);
	XIicPs_MasterSendPolled(InstancePtr, &buffer[2], 2, I2C_SI5341_ADDR);
}

#if 0
static void
si5341_i2c_write_nvm(XIicPs *iic)
{
	uint8_t buffer[4];

	// see p13 of the reference manual Si5341-40-D-RM

	xil_printf("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n\r");
	xil_printf("Si5341: wiriting to NVM!!!\n\r");

	// write 0xC7 to NVM_WRITE register
	buffer[0] = 0x01;
	buffer[1] = 0;
	buffer[2] = 0xE3;
	buffer[3] = 0xC7;

	XIicPs_MasterSendPolled(iic, buffer, 2, I2C_SI5341_ADDR);
	XIicPs_MasterSendPolled(iic, &buffer[2], 2, I2C_SI5341_ADDR);

	// wait until DEVICE_READY = 0x0F
	buffer[0] = 0x01;
	buffer[1] = 0;
	buffer[2] = 0xFE;
	XIicPs_MasterSendPolled(iic, buffer, 2, I2C_SI5341_ADDR);
	XIicPs_MasterSendPolled(iic, &buffer[2], 1, I2C_SI5341_ADDR);
	do {
		buffer[0] = 0;
		XIicPs_MasterRecvPolled(iic, buffer, 1, I2C_SI5341_ADDR);
	} while (*buffer != 0xF);

	// Set NVM_READ_BANK 0x00E4[0] = 1
	buffer[0] = 0x01;
	buffer[1] = 0;
	buffer[2] = 0xE4;
	buffer[4] = 1;
	XIicPs_MasterSendPolled(iic, buffer, 2, I2C_SI5341_ADDR);
	XIicPs_MasterSendPolled(iic, &buffer[2], 2,
			I2C_SI5341_ADDR);

	// wait until DEVICE_READY = 0x0F
	buffer[0] = 0x01;
	buffer[1] = 0;
	buffer[2] = 0xFE;
	XIicPs_MasterSendPolled(iic, buffer, 2, I2C_SI5341_ADDR);
	XIicPs_MasterSendPolled(iic, &buffer[2], 1,
			I2C_SI5341_ADDR);
	do {
		buffer[0] = 0;
		XIicPs_MasterRecvPolled(iic, buffer, 1, I2C_SI5341_ADDR);
	} while (*buffer != 0xF);

	xil_printf("Si5341: wiriting to NVM DONE!!!\n\r");
	xil_printf("////////////////////////////////\n\r");
}
#endif

int
init_si5341(void)
{
	XIicPs iic;
	XIicPs_Config *iic_config;
	int32_t status;

	/* Initialize the IIC0 driver so that it is ready to use */
	iic_config = XIicPs_LookupConfig(IIC_DEVICE_ID);
	if (iic_config == NULL) {
		xil_printf("XIicPs_LookupConfig error\r\n");
		return -1;
	}

	status = XIicPs_CfgInitialize(&iic, iic_config, iic_config->BaseAddress);
	if (status != XST_SUCCESS) {
		xil_printf("XIicPs_CfgInitialize error\r\n");
		return -1;
	}

	/* Set the IIC serial clock rate */
	status = XIicPs_SetSClk(&iic, IIC_SCLK_RATE);
	if (status != XST_SUCCESS) {
		xil_printf("XIicPs_SetSClk error\r\n");
		return -1;
	}

        xil_printf("I2C configured\r\n");

	// check if mux is alive
	status = IicPsSlaveMonitor(&iic, I2C_MUX_ADDR);
	if (status == XST_SUCCESS) {
		xil_printf("I2C mux alive\r\n");
	} else {
		xil_printf("I2C mux not alive\r\n");
		return -1;
	}
	i2cmux_set_channel(&iic, I2C_MUX_CH_SI5341);

	// send preamble
	for (int i = 0; i < 6; i++) {
		si5341_i2c_write(&iic, si5341_revd_registers[i].address,
				si5341_revd_registers[i].value);
	}

	xil_printf("Si5341: preamble sent\n\r");

	/* Delay 300 msec */
	usleep(300000);

	for (int i = 6; i < SI5341_REVD_REG_CONFIG_NUM_REGS; i++) {
		si5341_i2c_write(&iic, si5341_revd_registers[i].address,
				si5341_revd_registers[i].value);
	}
	xil_printf("Si5341: config sent\n\r");

	/* Wait until bus is idle */
	while (XIicPs_BusIsBusy(&iic) > 0)
		;

	/* Soft reset */
	si5341_soft_rst(&iic);
	usleep(300000);

	/* Wait for dev ready */
	si5341_dev_ready(&iic);

	xil_printf("Si5341: Completed!\n\r");

	/* Configure I2C Mux to SFP */
	i2cmux_set_channel(&iic, I2C_MUX_CH_SFP);

        xil_printf("I2C mux done\r\n");

        return 0;
}
